/*****************************************************************************/
/* File:         afferent.cpp                                                */
/* Version:      4.0                                                         */
/* Description:  methods for the afferent class (defined in basal_ganglia.h) */
/* References:                                                               */
/* Authors:      Tony Prescott                                               */
/* Contact:      T.j.prescott@shef.ac.uk                                     */
/*****************************************************************************/

#ifndef AFFERENT
#define AFFERENT

#include <stdio.h>
#include "nucleus.h"

namespace bgs
{

  class nucleus;
  class afferent;
  typedef vector <afferent*> avector;

  class afferent
    {
    private:
      nucleus* aNucleusF;      // pointer to the nucleus from which the connection arises
      nucleus* aNucleusT;      // pointer to the nucleus where the connection terminates
      float aWeight_w;         // generic weight on within-channel input
      float aWeight_b;         // generic weight on between-channel input on different channel
      float* aChannel_ws;      // specific weights for within channel input
      int aInput_type;
    public:
      afferent(nucleus*, nucleus*, float, float wb=0.0, float cws[]= NULL);
      int input_type();
      float weight_w() { return aWeight_w; }
      float set_weight_w(float v) { return aWeight_w= v; }
      float weight_b() { return aWeight_b; }
      float set_weight_b(float v) { return aWeight_b= v; }
      float weighted_input(int);
      float weighted_input0(int);
      float weighted_input(int, nucleus*);    // for debugging only
      void print(int);
    };

  // constructor

  afferent::afferent(nucleus* nf, nucleus* nt, float ww, float wb, float* cws)
  { 
    int nn, i;
    // nucleus *nptr= (nucleus*) nuc; 
    
    // cout << "new input from " << nuc->name() << endl;
    aNucleusF= nf;
    aNucleusT= nt;
    aWeight_w= ww; 
    aWeight_b= wb;
    aChannel_ws= NULL;  // added 07.08.03 fixes bug in Codewarrior version
    
    if (cws!=NULL)
      {
    cerr << "Warning: specific weights on afferent from " << nf->name() << endl; 
   	nn= aNucleusF->n_neurons();
	aChannel_ws= new float[nn];
	for (i=0; i<nn; ++i) aChannel_ws[i]= cws[i];
      } 
    aInput_type= input_type();
  }
  

  int afferent::input_type()
  {
    if (aWeight_w==0 && aWeight_b==0 && aChannel_ws==NULL) return -1;
    else if (aWeight_w!=0 && aWeight_b==0 && aChannel_ws==NULL) return 0;
    else if (aWeight_w==0 && aWeight_b!=0 && aChannel_ws==NULL) return 1;
    else if (aWeight_w!=0 && aWeight_w==aWeight_b && aChannel_ws==NULL) return 2;
    else if (aWeight_w!=0 && aWeight_b!=0 && aChannel_ws==NULL) return 3;
    else if (aWeight_w!=0 && aWeight_b==0 && aChannel_ws!=NULL) return 4;            //*** 4.3.03
    else if (aWeight_w==0 && aWeight_b==0 && aChannel_ws!=NULL) return 5;            //*** 4.3.03
    else return -2;
  }
  

  // KEY FUNCTION!!
  
  inline float afferent::weighted_input(int c)
  {
    float within, between= 0.0;
    float out_w= aNucleusF->output(c);
    
    // within channel input
    within= aWeight_w * out_w;             // generic within channel weight
    if (aChannel_ws!=NULL)                 // if there is a specific channel weight
      {
	// float s= aChannel_ws[c]*out_w;
	// if (s!=0.0) 
	//  cerr << "adding " << s << " on channel " << c << endl;
	within += aChannel_ws[c] * out_w;    // ADD that in   *** 27.01.03 was multiply!
      }	
    // between channel input
    if (aWeight_b!=0)
      between= aWeight_b * (aNucleusF->total_output() - out_w);
    // note: this is the same but much faster than the following
    // for (int i=0; i<aNucleus->n_neurons(); ++i)
    //  if (i!=c) between += aWeight_b * aNucleus->output(i);
    
    return within + between;
  }


  // an alternative way of doing the above (and about at the same speed)
  
  inline float afferent::weighted_input0(int c)
  {
    float out_w= aNucleusF->output(c);
    
    switch(aInput_type)
      {
      case -1: return 0.0;                                            // zero weights
      case 0: return aWeight_w * out_w;                               // within only
      case 1: return aWeight_b * (aNucleusF->total_output() - out_w); // between only
      case 2: return aWeight_w * aNucleusF->total_output();           // within=between
      case 3: return (aWeight_w * out_w) +                            // within!=between
		(aWeight_b * (aNucleusF->total_output() - out_w));
      // **** next line add not multiply
      case 4: return (aWeight_w + aChannel_ws[c]) * out_w;           // within + specific
      case 5: return aChannel_ws[c] * out_w;                         // specific
      default: 
	cout << "afferent: weighted input() unrecognised type\n";
	return 0.0;
      }
  }
  

  // The rest is for display/monitoring purposes
    
  void afferent::print(int verbose= 1)
  {
    int nn= aNucleusF->n_neurons();
    
    if (verbose==1) 
      {
	cout << aNucleusF->name();
	switch (aInput_type)
	  {
	  case -1: cout << "0"; break;
	  case 0: cout << ((aWeight_w>0)? "+" : "-"); break;
	  case 1: cout << ((aWeight_b>0)? ":+" : ":-"); break;
	  case 2: 
	  case 3:
	    cout << ((aWeight_w>0)? "+" : "-");
	    cout << ((aWeight_b>0)? ":+" : ":-"); break;
	  case 4: cout << ((aWeight_w>0)? "+$" : "-$"); break; 
	  case 5: cout << "$"; break;
	  default: cout << "??";
	  }
      }
    else if (verbose>1)
    {
      if (verbose>2) cout << aNucleusF->name() << "-" << aNucleusT->name() << ": ";
	  else cout << "  input: " << aNucleusF->name() << " ";
	switch (aInput_type)
	  {
	  case -1: cout << "inactive\n"; break;
	  case 0: cout << "within(" << aWeight_w << ")\n"; break;
	  case 1: cout << "between(" << aWeight_b << ")\n"; break;
	  case 2: cout << "within = between(" << aWeight_w << ")\n"; break;
	  case 3: cout << "within(" << aWeight_w << ") /= between(" << aWeight_b << ")\n"; break;
	  case 4: cout << "within: " << aWeight_w << " + ("; 
	  case 5:
	    if (aInput_type==5) cout << "within(";
	    for (int i=0; i<(nn-1); ++i) cout << aChannel_ws[i] << ", ";
	    cout << aChannel_ws[nn-1] << ")\n";
	    break;
	  default: 
	    cout << "(" << aWeight_w << ", " << aWeight_b << ") specific weights ";
	    if (aChannel_ws!=NULL) 
	      {
		cout << "(";
		for (int i=0; i<(nn-1); ++i) cout << aChannel_ws[i] << ", ";
		cout << aChannel_ws[nn-1] << ")\n";
	      }
	    else cout << "not defined\n";
	  }
      }
  }
  

  // for debugging purposes yet another version of weighted input
  float afferent::weighted_input(int c, nucleus* to)
  {
    int i;
    float within, between= 0.0;
    
    // within channel input
    within= aWeight_w * aNucleusF->output(c);    // generic within channel weight
    if (aChannel_ws!=NULL)                       // if there is a specific channel weight
      within *= aChannel_ws[c];                  // multiply that in                 
    
    // between channel input
    if (aWeight_b!=0)
      for (i=0; i<aNucleusF->n_neurons(); ++i)
	if (i!=c) 
	  between += aWeight_b * aNucleusF->output(i);
    
    printf("%s-%s(%3.1f%s, %3.1f): %d ", aNucleusF->name(), to->name(), 
	   aWeight_w, (aChannel_ws!=NULL)? "*" : "", aWeight_b, c);
    printf("%4.2f+%4.2f=%4.2f\n", within, between, within+between);
    
    return within + between;
  }
  

  // NUCLEUS METHODS INVOLVING AFFERENTS

	/*
  afferent* nucleus::add_input(nucleus *nuc, float ww, float wb)
    {
      if (ww!=0.0 || wb != 0.0)              // check for zero weights
		{
	  		afferent *a= new afferent(nuc, ww, wb); 
	  		aInputs.push_back(a); 
	  		return a; 
		}
      else return NULL;
    }
    */
	
  afferent* nucleus::add_input(nucleus *nuc, float ww, float wb, float cweights[])
  	// wb defaults to 0.0
  	// cweights defaults to NULL
    {
      if (ww!=0.0 || wb != 0.0 || cweights != NULL)              // check for zero weights
		{
	  		afferent *a= new afferent(nuc, this, ww, wb, cweights); 
	  		aInputs.push_back(a); 
	  		return a; 
		}
      else return NULL;
    }
  
  // *** two short-cut versions of add_input removed (because they could create confusion)

  // the following are all declared inline to maximise speed
  // KEY FUNCTION !!
  
  inline float nucleus::net_input(int channel, float salience)
    {
      float net= 0.0;
      
      if (aSalience_weight!=0.0) net+= aSalience_weight * salience;
      
      avector::iterator avi;
      for (avi= aInputs.begin(); avi!=aInputs.end(); ++avi)
	net+= (*avi)->weighted_input(channel);         // for debugging use (i, this)
      
      // the following does the same but is slightly slower
      // for (uint i=0; i<aInputs.size(); ++i)
      // net+= aInputs[i]->weighted_input(channel);
      
      // *** now doing Da modululation using da_leaky class, 24.03.03
      // if (aDa_modulation!=0) net*=(1.0 + (aDa_modulation * global_da));
      return net;
    }
  


}   // end of namespace


#endif


















