/****************************************************************************
* File:         basal_ganglia.h                                             *
* Version:      4.0                                                         *
* Description:  basal ganglia class and methods                             *
* Authors:      Tony Prescott and Fernando Montes                           *
* Contact:      T.j.prescott@shef.ac.uk                                     *
*****************************************************************************/

#ifndef BASAL_GANGLIA
#define BASAL_GANGLIA

#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include "../utils/utils.h"

enum  Da_type {eNone= 0, eD1= 1, eD2= 2, eD1f= 3, eD2f= 4};  // *** 
float global_da= 0.22;

const int n_sflags= 7;
char* sflag_names[n_sflags]= { "none", "part", "cln", "intf", "dual", "mult", "total"};

#include "leaky.h"           // class for leaky neuron
#include "da_leaky.h"        // class for dopamine modulated leaky neurons
#include "nucleus.h"         // class for a model nucleus (e.g. cortex, gpi, etc.)
#include "afferent.h"        // class of inputs to model nuclei


using namespace std;

namespace bgs
{
  class basal_ganglia :public data_log    // data_log defined in utils/data_log.h
    {
    private:
      string aName;
      string aVersion;
      string aDate;
      string aDescription;
      
      int aLog_level;                     // sets level of log file output
      
      int aN_channels;
      int aN_pchannels;                   // number of channels to print
      
      nvector aNuclei;
      nucleus *aCtx;
      nucleus *aSd1;
      nucleus *aGpi;
      nucleus *aThl;
      nucleus *aStn;

      avector aAfferents;                // afferent available for weight modification

      float aTime_step;
      float aStriatal_mi;
      float aOutput_mi;
      float aGpi_tonic;                  // *** new attribute 27.01.03
      bool  aSpecific_persist;
      float aTolerance;
      
      float aMaxd;
      int   aSteps;
      timer aTimer;                      // timer defined utils/timer.h  
     
      int   aWinner;
      int   aSflag;                      // selection outcome (none, part, clean, distortion, dual)
      int   aDraw;
      float aEfficiency;                 // efficiency of the selected behaviour (0-1.0)
      float aDistortion;                 // distortion of the selected behaviour (0-1.0)

      int   aSflag_count[n_sflags];
      
      stat aTime_stat;                   // stat defined utils/stat.h
      stat aConv_stat;
      stat aEff_stat;  
      stat aDis_stat;
      stat aWin_stat;
      stat aMarg_stat;
      bv_stat aS_surf;                     
      float aWin_s;                      // salience (cortical output) of winning channel
      float aNext_s;                     // salience of strongest loser
 
      float set_time_step(float f) { return aTime_step= f; }
      float set_striatal_mi(float f) { return aStriatal_mi= f; }
      float set_output_mi(float f) { return aOutput_mi= f; }
      bool  set_specific_persist(bool b, float* ws) { return aSpecific_persist= (b && ws!=NULL); }

    public:
      // inline methods
      void set_name(string nm, string vs, string dat) { aName= nm; aVersion= vs; aDate= dat; }
      void set_description(string ds) { aDescription= ds; }
      int set_log_level(int i) { return aLog_level= i; }
      string name() { return "Basal ganglia: "+aName; }
      int n_channels() { return aN_channels; }
      int n_nuclei() { return aNuclei.size(); }
      nvector& nuclei() { return aNuclei; }
      float time_step() { return aTime_step; }
      float maxd() { return aMaxd; }
      float dopamine() { return global_da; }
      float set_dopamine(float f) { global_da= f; return f; }
      float set_tolerance(float f) { return aTolerance= f; }
      float tolerance() { return aTolerance; }
      bool  specific_persist() { return aSpecific_persist; }
      float striatal_mi() { return aStriatal_mi; }
      float output_mi() { return aOutput_mi; }
      float gpi_tonic() { return aGpi_tonic; }
      afferent* get_afferent(int i) { return aAfferents[i]; }
      void add_afferent(afferent* a) { return aAfferents.push_back(a); }
      void print_afferents() { print_vector(aAfferents, 3); }

      nucleus* add_nucleus(nucleus* nptr) { aNuclei.push_back(nptr); return nptr; }
      
      // compute the gating signals by inverting the GPi output use Gpi_tonic as threshold
      void gates(float values[], bool talk = false) 
	{ if (talk) aGpi->print_values(); aGpi->invert_output(values, aGpi_tonic); }
      
      // use the thalamic output to provide feedback to behaviours
      void feedback(float values[]) { aThl->threshold_output(values); }

      // get the normalised saliences (cortical output)
      void cortex(float values[]) { aCtx->values(eY, values); }

      // transform salience (you provide) values into the range 0-1 using the Sd1 piecewise-lin method
      void pwl_salience(float values[]) { aCtx->pwl_values(values); }
      
      // building the basal ganglia
      basal_ganglia(int, float*, int);
      void build(float*);                          // defined in bg_build.h
      void zero();
      void calibrate(bool init= true);
      
      // running the model
      float step(float*);
      int converge(float*, int, bool, bool);

      // monitoring and output
      int find_winner();
      int winner() { return aWinner; }
      float win_sal() { return aWin_s; }
      float next_sal() { return aNext_s; }
      bv_stat* surface() { return &aS_surf; }
      int sflag() { return aSflag; }
      int draw() { return aDraw; }
      float efficiency() { return aEfficiency; }
      float distortion() { return aDistortion; }
      int   activity_steps() { return aSflag_count[n_sflags-1] - aSflag_count[0]; }
            // total steps - number of steps with no selection 

      void print_description(int);
      float* values(int, Leaky_value, float*);
      char* nucleus_name(int);
      void print(int);
      void print(int, string, bool);
      void print_array(char*, float*);
      void print_values(int, Leaky_value);
      void print_values(Leaky_value);
      void print_nstats();
      void print_analysis_hdr(bool, int);
      void print_analysis_r(bool, int);
      void print_analysis(int);
      virtual void log_header();
      virtual void log();
      void log(float*, int, int);
      void log(double);
    };
  
  
  
  // constructor

  basal_ganglia::basal_ganglia(int noc, float persist[], int nopc= 0) : 
    aN_channels(noc), aN_pchannels(nopc), aTimer("bg time", 3), 
    aTime_stat("BG time ms", false), 
    aConv_stat("convergence", false, true),
    aEff_stat("efficiency", false),
    aDis_stat("distortion", false),
    aWin_stat("Winner", false),
    aMarg_stat("Margin", false), // for distribution add: ,  true, 0.0, 0.25),
    aS_surf("Salience", 10)
    {
      aTime_step= 0.05;       // default time step
      aSpecific_persist= false;
      aTolerance= 0.0001;
      aGpi= NULL;
      if (aN_pchannels==0 || aN_pchannels>aN_channels) aN_pchannels= aN_channels;
      build(persist);
      calibrate();
    }
    
  // note: build, the main function for constructing the basal ganglia, is in BG_BUILD.H
  
  
  // Step- the main update function
  
    float basal_ganglia::step(float salience[])
    {
      float d;
      aMaxd= 0.0;
      
      // cout << "basal ganglia::Step- updating activations\n";
      for (unsigned i= 0; i<aNuclei.size(); ++i)
	{
	  d= aNuclei[i]->update_activation(salience);
	  if (d>aMaxd) aMaxd= d;
	}
      
      // cout << "basal_ganglia::Step- updating output\n";
      for (unsigned i= 0; i<aNuclei.size(); ++i) aNuclei[i]->update_output();
      
      return aMaxd;
    }
  
  
  // Converge- runs step until the activations are changing less that TOLERANCE
  
  int basal_ganglia::converge(float *salience, int epoch= -1, bool pflag= false, 
			                                            bool save= true)
    {
      const int max_steps= 1000;
      
      float max_diff0= 1.0, max_diff1= 1.0;
      int  i= 0;
      
      aSteps= max_steps+1;
      
      aTimer.start();
      while (i<max_steps && aSteps>max_steps)
	{
	  ++i;
	  max_diff1= step(salience);
	  if (max_diff1<aTolerance && max_diff0<aTolerance) 
	    aSteps= i - 1;
	  max_diff0= max_diff1;
	  if (save && record_level()>0) log(salience, epoch, i);
	}
      aTimer.stop();
      if (save)
	{
	  find_winner();
	  if (record_level()==0 && epoch>=0) log(salience, epoch, i);
	  
	  for (unsigned i= 0; i<aNuclei.size(); ++i) aNuclei[i]->update_stat();
	  
	  if (aSteps>1)
	    {
	      aTime_stat.add_value(aTimer.read());
	      aConv_stat.add_value(aSteps);
	    }
      
	  if (pflag && aSteps<=max_steps)
	    printf("bg converged in %d steps, max delta %10.8f, time taken %3.2fms\n", 
		   aSteps, max_diff1, aTimer.read());
	}
      if (aSteps>max_steps)
	printf("bg failed to converge in %d steps, max delta= %7.5f\n", max_steps, max_diff1);
      
      return aSteps;
    }
  
  void basal_ganglia::zero()
  {
     aWinner= aSflag= -1;
     for (int i= 0; i<n_sflags; ++i) aSflag_count[i]= 0;
     aEfficiency= aDistortion= 0.0;
     aTime_stat.zero();
     aConv_stat.zero();
     aEff_stat.zero();
     aDis_stat.zero();
     aWin_stat.zero();
     aMarg_stat.zero();
     aS_surf.zero();
     aWin_s= aNext_s= 0.0;
     for (unsigned i= 0; i<aNuclei.size(); ++i) aNuclei[i]->zero();
  }
  
  void basal_ganglia::calibrate(bool init)
    // init defaults to true
    {
      // cerr << "running calibration...\n";
      float* tmp= new float[aN_channels];
      for (int i=0; i< aN_channels; ++i) tmp[i]= 0.0;
      converge(tmp, -1, false, false);
      if (init) zero();
      // osprint(cout, aN_channels, aGpi->values(eY, tmp), ' ');
      delete[] tmp;
      aGpi_tonic= aGpi->value(eY, 0);
      // cout << "GPi tonic value= " << gpi_tonic() << endl;
      // aThl->set_thresholds(-aGpi_tonic);
    }
  
  int basal_ganglia::find_winner()
    {
      float* garr= new float[aN_channels];
      float* sarr= new float[aN_channels];

      gates(garr, false);
      cortex(sarr);             // cortical output is equal to normalised salience

      int win= -1, lose= -1, ws= -1;
      int n_part= 0, n_full= 0;
      float max_g= -1.0, next_g= -1.0, sum_g= 0.0;
      float win_s= -1.0, next_s= -1.0;
      float draw_tol= 0.01;
      
      aDraw= 0;
      for (int i= 0; i<aN_pchannels; ++i)
	{
	  float g= garr[i];
	  float s= sarr[i];
	  sum_g+= g;
	  if (g > max_g) 
	    { lose= win; win= i; next_g= max_g; max_g= g; }
	  else if (g > next_g) 
	    { next_g= g; lose= i; }

	  if (g >= 0.95) n_full++;
	  else if (g >= 0.05) n_part++;

	  if (s > win_s) { ws= i; next_s= win_s; win_s= s; }
	  else if (s > next_s) { next_s= s; }
	}
      if (ws!=win) { next_s= win_s; win_s= sarr[win]; }
      delete[] garr;
      delete[] sarr;

      if (n_full==0 && n_part==0) aSflag= 0;           // no selection
      else if (n_full==0) aSflag= 1;                   // partial selection
      else if (n_full==1 && n_part==0) aSflag= 2;      // clean selection
      else if (n_full==1) aSflag= 3;                   // intereference
      else if (n_full==2) aSflag= 4;                   // dual selection
      else if (n_full>2) aSflag= 5;                    // multiple selection
      else cerr << "bg::find_winner, this shouldn't happen!\n";
      
      aEfficiency= max_g;
      // aDistortion= next_g;
      // aDistortion= sum_g - max_g;
      aDistortion= (sum_g==0.0)? 0.0: (sum_g - max_g)/sum_g;
      if (aSflag>0)
	{
	  aEff_stat.add_value(aEfficiency);
	  aDis_stat.add_value(aDistortion);
	}
      if ((max_g - next_g) <= draw_tol) aDraw++;
      
      aSflag_count[n_sflags - 1]++;                 // increment total flag count
      if (aSflag>=n_sflags - 1) 
	cerr << "bg::find_winner, invalid flag!\n";
      else aSflag_count[aSflag]++;                  // increment current flag
      aWinner= (max_g<=0)? -1 : win;
      aWin_stat.add_value(win_s);
      aMarg_stat.add_value(fabs(win_s - next_s));
      // cerr << "find_winner: ";
      // cerr << win << " " << lose << " " << max_s << " " << (max_s - next_s) << "\n";
      aS_surf.add_value(win_s, next_s);
      aWin_s= win_s; aNext_s= next_s;

      return aWinner;
    }
  
  // Monitoring and Output  
  
  void basal_ganglia::print_description(int verbose)
    {
      // set the description string for the model
      string ts, smis, omis, ds, gpis;
      ds= "da (" + ftos(dopamine()) + ")";
      gpis= " GPi tonic (" + ftos(gpi_tonic(), 3) + ")";
      if (striatal_mi() != 0.0) smis= " + str mi (" + ftos(striatal_mi(), 1) + ")";
      if (output_mi()) omis= " + gp mi (" + ftos(output_mi(), 1) + ")";
      if (specific_persist()) ts= " + specific persist";
      set_description(ds + gpis + smis + omis + ts);
      
      if (verbose==0)
	{
	  cout << "Basal ganglia: " << aName << " " << aVersion << ", updated " << aDate;
	  cout << "\n               " << aDescription << "\n";
	}
      else 
	{
	  cout << "\nBasal Ganglia Model\n";
	  cout << "Version: " << aName << " " << aVersion << ", updated " << aDate << "\n";
	  cout << "Description: " << aDescription << "\n";
	}
    }
  
  void basal_ganglia::print(int verbose= 1)
    {
      print_description(verbose);
      
      if (verbose>0)
	{
	  cout << aName << "(" << n_nuclei() << ", " << n_channels() << "):   ";
	  if (verbose==1) 
	    {
	      pprint_vector <nucleus*> (aNuclei);
	    }
	  else
	    {
	      cout << "\n";
	      for (unsigned i=0; i<aNuclei.size(); ++i) aNuclei[i]->print(verbose);
	    }
	  cout << endl;
	}
    }
 
  void basal_ganglia::print(int verbose, string fname, bool append= false)
    {
      ofstream outs;
      if (redirect_cout(outs, fname, append)) {
	print(verbose);
	reinstate_cout(outs);
      }
    }

  char* basal_ganglia::nucleus_name(int n) 
    {
      if (n < (int) aNuclei.size()) return aNuclei[n]->name();
      else {
	// printf("basal_ganglia::nucleus_name: attempt to access nucleus %d of %d\n", 
	//  	     n, aNuclei.size());
	return "";
      }
    }
    
  void basal_ganglia::print_analysis_hdr(bool el= true, int verbose= 0)
  {
    // cerr << "verbose is " << verbose << endl;
    cout << "DA\ttotal\t";
    for (int i=0; i<(n_sflags - 1); ++i) cout << sflag_names[i] << '\t';
    aCtx->nstat()->print_stats_hdr(false, verbose);
    aEff_stat.print_stats_hdr(false, verbose);
    aDis_stat.print_stats_hdr(false, verbose);
    aWin_stat.print_stats_hdr(false, verbose);
    aMarg_stat.print_stats_hdr(false, verbose);
    if (el) cout << endl; else cout << '\t';
  }
  
  void basal_ganglia::print_analysis_r(bool el= true, int verbose= 0)
  {
    int last= n_sflags - 1;

    cout << global_da << '\t';
    cout << aSflag_count[last] << '\t';
    for (int i=0; i<last; ++i)
      cout << 100.0*aSflag_count[i]/aSflag_count[last] << '\t';
    aCtx->nstat()->print_stats_r(false, verbose);
    aEff_stat.print_stats_r(false, verbose);
    aDis_stat.print_stats_r(false, verbose);
    aWin_stat.print_stats_r(false, verbose);
    aMarg_stat.print_stats_r(false, verbose);
    // aWin_stat.print(true, 100);
    if (el) cout << endl;
  }
  
  void basal_ganglia::print_analysis(int verbose= 0)
    {
      int ts= aSflag_count[n_sflags - 1];

      cout << "total steps:    " << ts << endl;
      for (int i=1; i<(n_sflags - 1); ++i) 
	cout << "bg-" << sflag_names[i] << ":   " << aSflag_count[i]/ts << "%\n";
      aEff_stat.print_stats();
      aDis_stat.print_stats();
      aTime_stat.print_stats();
      aConv_stat.print_stats();
    }

  void basal_ganglia::print_array(char* aname, float* a) 
    {
      printf("%s: ", aname);
      for (int i=0; i<aN_channels; ++i) printf("%7.4f ", a[i]);
      printf("\n");
    }
  
  float* basal_ganglia::values(int n, Leaky_value lv, float *values)
    {
      if (n<(int) aNuclei.size()) aNuclei[n]->values(lv, values);
      else printf("basal_ganglia::values: attempt to access unknown nucleus %d\n", n);
      return values;
    }     
  
  void basal_ganglia::print_values(int n, Leaky_value v) 
    {
      float* tmp= new float[aN_channels];
      
      values(n, v, tmp);
      print_array(nucleus_name(n), tmp);
      delete[] tmp;
    }
  
  void basal_ganglia::print_values(Leaky_value v)
    {
      for (unsigned i= 0; i<aNuclei.size(); ++i) aNuclei[i]->print_values(v);
      cout << endl;
    } 

  void basal_ganglia::print_nstats()
    {
      for (unsigned i= 0; i<aNuclei.size(); ++i) aNuclei[i]->nstat()->print_stats();
      cout << endl;
    } 
  
  // the following is called from log(int, int) or log(double), see data_log.h

  void basal_ganglia::log_header()
    {
      ofstream& outs= log_stream();

      if (aLog_level==0)
	{
	  outs << "t\tconv\twin\tflag\teff\ttstn\t";
	  // for (int j= 0; j<aN_pchannels; ++j) outs << "sal-" << (j+1) << " ";
	  // outs << " ";
	  for (unsigned i= 0; i<aNuclei.size(); ++i)
	    if (aNuclei[i]->save())
	      {
		for (int j= 0; j<aN_pchannels; ++j)
		  outs << aNuclei[i]->name() << "-" << (j+1) << '\t';
	      }
	  outs << endl;
	}
    }
	  
  void basal_ganglia::log()
    {
      ofstream& outs= log_stream();
      float* tmp= new float[aN_channels];

      outs << aWinner+1 << '\t' << aSflag << '\t' << aEfficiency << '\t';
      outs << aStn->total_output() << '\t';

      if (aLog_level==0)
	{
	  for (unsigned i= 0; i<aNuclei.size(); ++i)
	    if (aNuclei[i]->save())
	      osprint(outs, aN_pchannels, aNuclei[i]->values(eY, tmp), '\t');
	}
      else
	{
	  for (unsigned i= 0; i<aNuclei.size(); ++i)
	    {
	      outs << endl;
	      outs << aNuclei[i]->name() << ":" << '\t';
	      osprint(outs, aN_pchannels, aNuclei[i]->values(eY, tmp), '\t');
	    }
	}
      outs << endl;
      delete[] tmp;
    }

  void basal_ganglia::log (float* salience, int epoch, int step) // called by benchmark tests
    {
      ofstream& outs= log_stream();
      if (record_log() && outs) {
	outs << epoch << '\t' << step;
	if (aLog_level==0) outs << '\t';
	else outs << endl;
	// if (aLog_level>0) outs << "sal: ";
	// osprint(outs, aN_pchannels, salience, ' ');
	log();
      }	
    }
  
  void basal_ganglia::log(double et)               // called by controller
    {
      ofstream& outs= log_stream();
      if (record_log() && outs) {
	outs << et << '\t' << aSteps << '\t';
	log();
      }	
    }

}  // end of namespace

#endif













