/*****************************************************************************
 * File:         bg/bg_basic_f.h                                               *
 * Version:      4.0                                                         *
 * Description:  Method for customising basal ganglia model                  *
 * Authors:      Tony Prescott                                               *
 * Contact:      t.j.prescott@shef.ac.uk                                     *
 *****************************************************************************/

// *** MANY CHANGES IN WEIGHTS ETC. FROM PRIOR VERSIONS ***
 
#ifndef BG_BUILD
#define BG_BUILD
#include <string>
#include "../utils/utils.h"

namespace bgs
{

  void basal_ganglia::build(float thl_ctx_ws[])
  {

    // set the model name etc.
    set_name("Basic (new DA)", "5.0", "Date");

    // various key parameters
    // note: you can use the test_bg/step program to calculate the gpi tonic
 
    set_time_step(0.05);                       // time step of leaky integrators
    set_dopamine(0.2);                         // initial dopamine level
    set_striatal_mi(0.0);                      // level of striatal mutual inhibition, <=0
    set_output_mi(0.0);                        // level of GPe, GPi mutual inhibition, <=0 (-.2)
    set_specific_persist(false, thl_ctx_ws);   // allow channel-specific thalamocortical weights
                                               // true only if 'true' and thl_ctx_ws!=NULL 
    set_tolerance(0.0001);                     // criteria for convergence of the bg model

    const float thl_ctx= 0.0;                  // global thl_ctx 'persistence' weight
    const float stn_out= 0.9;                  // MH uses 0.8, KG 0.9
    const float gpe_gpi= -0.3;                 // MH uses -0.4, KG -0.3
    
    // *** new_da stuff
    const float d1_pivot= 0.5;                 // pivot point for D1 dopamine
    const float d2_gain= 0.8;                  // gain parameter for D2 dopamine
    // ***
  
    // create nuclei, parameters are 
    // name, log (0 or 1), number of channels, slope, threshold, and DA type (default none)
    nucleus* ctx= add_nucleus(new nucleus("ctx", 1, n_channels(), time_step(), 1.0, 0.0));
    // *** new_da stuff
    nucleus* sd1= add_nucleus(new nucleus("sd1", 1, n_channels(), time_step(), 1.0, 0.1, 
					  eD1f, 1.0, d1_pivot));
    nucleus* sd2= add_nucleus(new nucleus("sd2", 1, n_channels(), time_step(), 1.0, 0.1, 
					  eD2f, d2_gain));
    // ***
    nucleus* stn= add_nucleus(new nucleus("stn", 0, n_channels(), time_step(), 1.0, -0.25));
    nucleus* gpe= add_nucleus(new nucleus("gpe", 0, n_channels(), time_step(), 1.0, -0.2));
    nucleus* gpi= add_nucleus(new nucleus("gpi", 1, n_channels(), time_step(), 1.0, 
					  ((output_mi()!=0.0)? -0.19 : -0.2)));
    nucleus* thl= add_nucleus(new nucleus("thl", 1, n_channels(), time_step(), 1.0,  -gpi_tonic()));  
    
    // add the between-nuclei connections

    ctx->add_salience_input(1.0);

    if (specific_persist())                       // specific thl-ctx weight on each channel
      ctx->add_input(thl, 0.0, 0.0, thl_ctx_ws);  // plus global thl-ctx weight
    else ctx->add_input(thl, thl_ctx);            // global thl-ctx weight only

    sd1->add_input(ctx, 1.0);                     // CTX-SD1 with-chan excitation
    sd1->add_input(sd1, 0.0, striatal_mi());      // SD1 mutual inhibition (between channels only)
    
    sd2->add_input(ctx, 1.0);                     // CTX-SD2 with-chan excitation
    sd2->add_input(sd2, 0.0, striatal_mi());      // SD2 mutual inhibition (between channels only)
    
    stn->add_input(ctx, 1.0);                     // CTX-STN with-chan excitation
    stn->add_input(gpe, -1.0);                    // GPE-STN with-chan inhibition (-1.0)
    
    gpe->add_input(sd2, -1.0);                    // SD2-GPE with-chan inhibition (-1.0)
    gpe->add_input(stn, stn_out, stn_out);        // STN-GPE within AND between excitation
    gpe->add_input(gpe, 0.0, output_mi());        // GPE mutual inhibition (between channels only)
    
    gpi->add_input(sd1, -1.0);                    // SD1-GPI with-chan inhibition (-1.0)
    gpi->add_input(gpe, gpe_gpi);                 // GPE-GPI with-chan inhibition
    gpi->add_input(stn, stn_out, stn_out);        // STN-GPI within AND between excitation
    gpi->add_input(gpi, 0.0, output_mi());        // GPI mutual inhibition (between channels only)
    
    // thl->add_input(ctx, 1.0);                  // CTX-THL with-chan excitation (+1.0)
    thl->add_input(gpi, -1.0);                    // GPI-THL with-chan inhibition (-1.0)
    
    aCtx= ctx;
    aSd1= sd1;
    aGpi= gpi;
    aThl= thl;
  }

}  // end of namespace


#endif





