/*****************************************************************************
 * File:         bg/bg_extended.h                                            *
 * Version:      4.0                                                         *
 * Description:  Method for customising basal ganglia model                  *
 * Authors:      Tony Prescott                                               *
 * Contact:      t.j.prescott@shef.ac.uk                                     *
 *****************************************************************************/

// This version has salience input to the cortex rather than the striatum
 
#ifndef BG_BUILD
#define BG_BUILD
#include <string>
#include "../utils/utils.h"

namespace bgs
{

  void basal_ganglia::build(float thl_ctx_ws[])
    // thl_ctx_ws is a vector of weights 
    // s_scale globally scales the salience inputs
  {

    // set the model name
    set_name("Extended (new DA)", "5.0", "Date");

    // various key parameters
    // note: you can use the test_bg/step program to calculate the gpi tonic
 
    set_time_step(0.05);                       // time step of leaky integrators
    set_dopamine(0.3);                         // initial dopamine level
    set_striatal_mi(0.0);                      // level of striatal mutual inhibition, <=0
    set_output_mi(0.0);                        // level of GPe, GPi mutual inhibition, <=0 (-.2)
    set_specific_persist(false, thl_ctx_ws);   // allow channel-specific thalamocortical weights
                                               // true only if 'true' and thl_ctx_ws!=NULL 
    set_tolerance(0.0001);                     // criteria for convergence of the bg model 

    const float stn_out= 0.9;                  // MH uses 0.8, KG 0.9
    const float gpe_gpi= -0.3;                 // MH uses -0.4, KG -0.3

    // *** new_da stuff
    const float sd_t= 0.15;                    // threshold of striatal neurons
    const float d1_pivot= 0.2;                 // pivot point for D1 dopamine
    const float d2_gain= 0.8;                  // gain parameter for D2 dopamine
    // ***

    // create nuclei, parameters are 
    // name, log (0 or 1), number of channels, slope, threshold, and DA type (default none)
    nucleus* ctx= add_nucleus(new nucleus("ctx", 1, n_channels(), time_step(), 1.0,   0.0));
    nucleus* mot= add_nucleus(new nucleus("mot", 0, n_channels(), time_step(), 1.0,   0.0));
    // *** new_da stuff
    nucleus* sd1= add_nucleus(new nucleus("sd1", 1, n_channels(), time_step(), 1.0,   sd_t, 
					  eD1f, 1.0, d1_pivot));
    nucleus* sd2= add_nucleus(new nucleus("sd2", 1, n_channels(), time_step(), 1.0,   sd_t, 
					  eD2f, d2_gain));
    // ***
    nucleus* stn= add_nucleus(new nucleus("stn", 0, n_channels(), time_step(), 1.0,   -0.25));
    nucleus* gpe= add_nucleus(new nucleus("gpe", 0, n_channels(), time_step(), 1.0,  -0.2));
    nucleus* gpi= add_nucleus(new nucleus("gpi", 1, n_channels(), time_step(), 1.0, 
					  ((output_mi()!=0.0)? -0.19 : -0.2)));
    nucleus* thl= add_nucleus(new nucleus("thl", 1, n_channels(), time_step(), 1.0,   0.0));      
    
    // trn
    nucleus* trn= add_nucleus(new nucleus("trn", 0, n_channels(), time_step(), 1.0,   0.0)); 
                                     
    // add the between-nuclei connections

	afferent *a0, *a1;

    ctx->add_salience_input(1.0);

    mot->add_input(ctx, 1.0);                     // CTX-MOT with-chan excitation (1.0)
    if (specific_persist())                       // specific thl-mot weight on each channel
      mot->add_input(thl, 0.0, 0.0, thl_ctx_ws);  
    else                                          // global thl-ctx weight only (1.0)
      a0= mot->add_input(thl, 1.0);

    sd1->add_input(ctx, 0.5);                     // CTX-SD1 with-chan excitation 0.5*(1+DA)
    sd1->add_input(mot, 0.5);                     // MOT-SD1 with-chan excitation 0.5*(1+DA)
    sd1->add_input(sd1, 0.0, striatal_mi());      // SD1 mutual inhibition (between channels only)
    
    sd2->add_input(ctx, 0.5);                     // CTX-SD2 with-chan excitation 0.5*(1-DA)
    sd2->add_input(mot, 0.5);                     // MOT-SD2 with-chan excitation 0.5*(1-DA)
    sd2->add_input(sd2, 0.0, striatal_mi());      // SD2 mutual inhibition (between channels only)

    stn->add_input(ctx, 0.5);                     // CTX-STN with-chan excitation (0.5)
    stn->add_input(mot, 0.5);                     // MOT-STN with-chan excitation (0.5)
    stn->add_input(gpe, -1.0);                    // GPE-STN with-chan inhibition (-1.0)
    
    gpe->add_input(sd2, -1.0);                    // SD2-GPE with-chan inhibition (-1.0)
    gpe->add_input(stn, stn_out, stn_out);        // STN-GPE within AND between excitation
    gpe->add_input(gpe, 0.0, output_mi());        // GPE mutual inhibition (between channels only)
    
    gpi->add_input(sd1, -1.0);                    // SD1-GPI with-chan inhibition (-1.0)
    gpi->add_input(gpe, gpe_gpi);                 // GPE-GPI with-chan inhibition
    gpi->add_input(stn, stn_out, stn_out);        // STN-GPI within AND between excitation
    gpi->add_input(gpi, 0.0, output_mi());        // GPI mutual inhibition (between channels only)
        
    // extended model
    thl->add_input(mot, 1.0);                     // MOT-THL with-chan excitation (+1.0)
    thl->add_input(gpi, -1.0);                    // GPI-THL with-chan inhibition (-1.0)
    a1= thl->add_input(trn, -0.1, -0.4);            // TRN-THL with-chan (-0.125) between (-0.4) 
    // thl->add_input(trn, -0.1, -0.7);           // use for MH benchmarks

    trn->add_input(mot, 1.0);                     // CTX-TRN with-chan excitation (+1.0)
    trn->add_input(thl, 1.0);                     // THL-TRN with-chan excitation (+1.0)
    trn->add_input(gpi, -0.2);                    // GPI-TRN with-chan inhibition (-1.0)
    // trn->add_input(trn, 0.0, -0.0);               // TRN-TRN mutual inhibition (-0.2)

    aCtx= ctx;
    aStn= stn;
    aSd1= sd1;
    aGpi= gpi;
    aThl= thl;
    add_afferent(a0);
    add_afferent(a1);
  }

}  // end of namespace


#endif






