/****************************************************************************
* File:         da_leaky.h                                                  *
* Version:      4.0                                                         *
* Description:  da modulation for leaky integrator neurons                  *
* Authors:      Mark Humphries and Tony Prescott                            *
* Contact:      t.j.prescott@shef.ac.uk                                     *
*****************************************************************************/

#ifndef DA_LEAKY
#define DA_LEAKY


/*
These methods implement slope and output 
calculations for D1 and D2 striatal units. The parameter 
values contained within them ensure that normal selection and 
switching should occur in the *intrinsic* model with a dopamine level of 
0.5. Using these values in the TRN/extended model will result in 
multiple channel selection, so alternatives are suggested in the code. 
Note that the TRN model is more robust to changes in the pivot point, 
threshold, and dopamine level (though I haven't had time to quantify 
this yet).
*/

namespace bgs
{
 
  class da_leaky : public leaky {
  private:
    Da_type aDa_type;             // type of dopamine modulation
    int     aDa_sign;             // direction of DA modulation (+1 or -1)
  public:
    da_leaky(Da_type dt) : leaky(), aDa_type(dt) {}
    virtual void  init(float, float, float);
    virtual float integrate(float);
    virtual void print(int);
  };

  // int dcount= 0;

  void da_leaky::init(float ts, float slope, float thresh) 
    {
      leaky::init(ts, slope, thresh);
      if (aDa_type==eD1) aDa_sign= 1;
      else if (aDa_type==eD2) aDa_sign= -1;
      else aDa_sign= 0;
      // cout << "initialised da_leaky: " << ++dcount << endl;
    }

  inline float da_leaky::integrate(float net)
  {
    return leaky::integrate(net*(1.0 + (aDa_sign * global_da)));
  }

  // monitoring and output

  void da_leaky::print(int verbose= 1)
  {
    leaky::print(verbose);
    static char* da_labels[]= {"NO DA", "DA", "DA"};
    if (verbose>0)
      cout << ", " << da_labels[(int) aDa_type] << " " << aDa_sign * global_da;
  }


  class da_full_leaky : public leaky {
  private:
    Da_type aDa_type;             // type of dopamine modulation
    float   aDa_gain;             // could be changed to increase slope change with increased DA
                                  // includes direction (+ or -)
    float   aPivot;
    float   aBase_slope;          
  public:    
    da_full_leaky(Da_type dt) : leaky(), aDa_type(dt) {}
    virtual void init(float, float, float, float, float);
    virtual float piecewlin();
    virtual float piecewlin(float);
    virtual void print(int);
  };
  
  void da_full_leaky::init(float ts, float slope, float thresh, 
			   float gain= 1.0, float pivot= 0.0) 
    {
      leaky::init(ts, slope, thresh);
      aBase_slope= slope;
      aPivot= pivot;
      if (aDa_type== eD1f) aDa_gain= gain; 
      else if (aDa_type== eD2f) aDa_gain= -gain;
      else aDa_gain= 0;
    }
  
  inline float da_full_leaky::piecewlin()
    // assigns and returns the output aY to the Piecewise linear 
    // transform of the activation aA
    {        
      aSlope = aBase_slope + (aDa_gain * global_da);
      if (aSlope < 0) 
	{
	  cerr << "Warning!!! negative slope for da_leaky neuron (da_leaky.h)\n";
	  return aY= 0.0;
	}
      
      // D1 version of the ramp pivots around the "pivot" point 
      else if (aDa_type==eD1f) 
	{	
	  float denom = 1.0 / aPivot;
	  aY = aSlope * (aA - aThreshold) + ((1 - aSlope)/denom);
	} 
      else if (aDa_type==eD2f)
	aY = aSlope * (aA - aThreshold);
      else
	{
	  cerr << "Invalid Da_type " << (int) aDa_type << " in da_full_leaky::piecewlin()\n";
	  return aY= 0.0;
	}

      // rectify output...
      if (aY < 0.0) return aY = 0.0;
      else if (aY > 1.0) return aY = 1.0;
      else return aY;
    }
  
  inline float da_full_leaky::piecewlin(float v)
    // assigns and returns the output aY to the Piecewise linear 
    // transform of the activation aA
    {        
      float y;
      aSlope = aBase_slope + (aDa_gain * global_da);
      
      if (aSlope < 0) 
	{
	  cerr << "Warning!!! negative slope for da_leaky neuron (da_leaky.h)\n";
	  return 0.0;
	}
      
      // D1 version of the ramp pivots around the "pivot" point 
      else if (aDa_type==eD1) 
	{	
	  float denom = 1.0 / aPivot;
	  y = aSlope * (v - aThreshold) + ((1 - aSlope)/denom);
	}
      else
	y = aSlope * (v - aThreshold);
      
      // rectify output...
      if (y < 0.0) return 0.0;
      else if (y > 1.0) return 1.0;
      else return y;
    }

  void da_full_leaky::print(int verbose= 1)
    {
      leaky::print(verbose);
      static char* da_labels[]= {"NO DA", "DA", "DA", "D1", "D2"};
      if (verbose>0)
	{
	  cout << ", " << da_labels[(int) aDa_type];
	  cout << "[" << global_da << ", " << aDa_gain;
	  if (aDa_type==eD1f) cout << ", " << aPivot;
	  cout << "]";
	}
    }

} // end of namespace

#endif




  // D1
  // example values for intrinsic model
  // aThresh = 0.1;
  // aBase_slope = 1;		// make identical to D2 base slope....
  // aDa_gain = 1;		// could be changed to increase slope change with increased dopamine
  // aPivot = 0.2;

  // alternative values for TRN model
  // aThresh = 0.15;
  // aPivot = 0.5;
  
  // D2
  // example values for intrinsic model
  // aThresh = 0.1;
  // aBase_slope = 1;	// make identical to D1 base slope....
  // aDa_gain = 0.8;		// ensures that output at DA = 1 is not zero - can change this

  // alternative values for TRN model
  // aThresh = 0.15;









