/****************************************************************************
* File:         leaky.h                                                     *
* Version:      4.0                                                         *
* Description:  leaky neuron class                                          *
* Authors:      Tony Prescott and Fernando Montes                           *
* Contact:      t.j.prescott@shef.ac.uk                                     *
*****************************************************************************/

#ifndef LEAKY
#define LEAKY

#define TEST2 0.2
#define DEFAULT_TIME_STEP  0.05         // time-step for the leaky integrators
#define STEP_SIZE_FRACT    0.3          // *** multiplied by TS to get step size 
                                        // should be < 0.5 to avoid instability 

namespace bgs
{

  enum    Leaky_value {eU= 0, eA= 1, eY= 2};
  class   leaky;
  typedef vector <leaky*> lvector;

  class leaky {
   protected:
    float aTime_step;
    float aK;
    float aStep;
    
    float aSlope;
    float aThreshold;
    
    float aU;
    float aA;
    float aY;
    float aD;
   public:
    float net() {return aU;}
    float activity() {return aA;}
    float output() {return aY;}
    float delta() {return aD; }
    float set_activity(float new_A) {aA= new_A; return aA;}
    
    leaky();
    virtual void  init(float, float, float);
    virtual void  init(float ts, float sl, float th, float gn, float pt) {}
    virtual float integrate(float);
    virtual float piecewlin();
    virtual float piecewlin(float);
    float set_threshold(float);
    float invert_output(float);
    float threshold_output(float);   // *** changed 27.01.03
    float value(Leaky_value);
    virtual void  print(int);
  };
    
  // constructor and initialisation

  // int lcount= 0, icount= 0;

  leaky::leaky()
  {
    aTime_step= DEFAULT_TIME_STEP;
    aK= 1/aTime_step;
    aStep= 0.3*aTime_step;           // must be 0.5* or less
    
    aSlope= 1.0;
    aThreshold= 0.0;
    aU= 0.0; aA= 0.0; aY= 0.0; 
    // cout << "created leaky " << ++lcount << endl;
  }
  
  void leaky::init(float ts, float slope, float thresh) 
  { 
    aTime_step= ts;
    aK= 1/ts;
    aStep= 0.3*ts;                   // must be 0.5* or less
    
    aSlope= slope;
    aThreshold= thresh;
    aU= 0.0; aA= 0.0; aY= 0.0; 
    // cout << "initialised leaky " << ++icount << endl;
  }
  

  // the following are all declared inline for speed

  inline float leaky::integrate(float net)
  {
    aU= net;
    aD= (-aK * (aA - aU)) * aStep;
    aA+= aD;
    return fabs(aD);
  }
  
  inline float leaky::piecewlin()
    // assigns and returns the output aY to the Piecewise linear 
    // transform of the activation aA
  {  
    if (aA < aThreshold) return aY= 0.0;
    else if (aThreshold <= aA && aA <= (1/aSlope + aThreshold)) 
      return aY= aSlope * (aA - aThreshold);
    else return aY= 1.0;
  }     
  
  inline float leaky::set_threshold(float thresh)
    {
      return aThreshold= thresh;
    }

  inline float leaky::piecewlin(float v)
    // returns the Piece-wise linear transform of the parameter v
  {  
    if (v < aThreshold) return 0.0;
    else if (aThreshold <= v && v <= (1/aSlope + aThreshold)) 
      return aSlope * (v - aThreshold);
    else return 1.0;
  }  
  
  inline float leaky::invert_output(float thresh= -99)
    // *** changed 27.01.03 to use thresh value passed as a parameter value
    // converts outputs in the range 0.0 to THRESH to the range 1.0 to 0.0
    // used for converting GPi outputs into a motor gating signal
  {
    if (thresh == -99) thresh= -aThreshold;
    float a= 1 - (1/thresh)*aY;
    
    if (a < 0.0) return 0.0;
    else if (0.0 <= a && a <= 1.0) return a;
    else return 1.0;
  }
  
  inline float leaky::threshold_output(float offset= 0)
    // converts output values equal to or below the neuron threshold
    // (-aThreshold) to 0.  Used to compute feedback values from thalamic output.
  {
    float a= aY - ((-aThreshold) + offset);
    
    if (a<0.0) return 0.0;
    else return a;
  }  
  
  inline float leaky::value(Leaky_value v)
  {
    switch(v){
    case eU: return aU;
    case eA: return aA;
    case eY: return aY;
    default:
      cout << "leaky.value-This shouldn't happen" << endl; return -1;
    }
  }


  // monitoring and output

  void leaky::print(int verbose= 1)
  {
    if (verbose>0)
      cout << aSlope << ", " << aThreshold << ", " <<  aTime_step << ", " << aStep;
    if (verbose>1)
      cout << " (U " << aU << ", A " << aA << ", Y " << aY << ")\n";
  }
 
}  // end of namespace

#endif











