/****************************************************************************
* File:         nucleus.cpp                                                 *
* Version:      4.0                                                         *
* Description:  nucleus class and methods                                   *
* References:                                                               *
* Authors:      Tony Prescott                                               *
* Contact:      T.j.prescott@shef.ac.uk                                     *
*****************************************************************************/

#ifndef NUCLEUS
#define NUCLEUS

#include <stdio.h>
#include "leaky.h"           // class for leaky neuron
#include "da_leaky.h"

namespace bgs
{

  class nucleus;
  class afferent;
  typedef vector <afferent*> avector;
  typedef vector <nucleus*> nvector;


  class nucleus
    {
    private:
      char* aName;     
      int   aSave;                               // 0- no log, 1- save to log
      int   aN_neurons;                          // number of leaky neurons
      float aSlope;                              // slope (M) for the leaky neurons
      float aThreshold;                          // threshold (T) for the leaky neurons
      float aSalience_weight;                    // weighting, if any, on salience inputs (pre-DA)
      lvector aNeurons;                          // vector of leaky integrators ***
      float aMaxd;                               // maximum change on last update_activation
      float aTotal_output;                       // sum of outputs from all neurons
      avector aInputs;                           // vector of afferent inputs
      // avector is defined in afferent.h
      stat  aStat;                               // stat is defined in utils/stat.h
    public:
      virtual int n_neurons() { return aN_neurons; }
      virtual char* name() { return aName; }
      int save() { return aSave; }
      virtual float output(int c) { return aNeurons[c]->output(); }
      virtual float total_output() { return aTotal_output; }
      float value(Leaky_value lv, int c) { return aNeurons[c]->value(lv); }
      leaky* neuron(int i) { return aNeurons[i]; }
      float salience_weight() { return aSalience_weight; }
      float add_salience_input(float w= 1.0) { return aSalience_weight= w; }
      stat*  nstat() { return &aStat; }
      void zero() { aStat.zero(); }
      void set_thresholds(float);
      virtual ~nucleus() {}                     // should probably do something!
    
      nucleus(char*, int, int, float, float, float, Da_type, float, float); // constructor
      float update_activation(float*);          // update the activation of the leaky integrators   
      void update_output();                     // update the outputs
      void update_stat();
      float* values(Leaky_value, float*);       // leaky values on all channels
      float* invert_output(float*, float thresh= -99);      // *** change in param list 27.03.03
      float* threshold_output(float*, float offset= 0);
      float* pwl_values(float*);
      void print(int);
      void print_values(Leaky_value lv);
      
      // the following methods are defined in afferent.h
      afferent* add_input(nucleus*, float, float wb= 0.0, float* csw= NULL); // adding inputs
                                                                
      // *** two short-cut versions removed 27.01.03
      float net_input(int, float);              // calculate the net input to a leaky integrator
    };
  

  // constructor
  
  nucleus::nucleus(char* name, int save, int n_neurons, 
		   float time_step, float slope, float threshold,
		   Da_type dt= eNone, float da_gain= 1.0, float da_pivot= 0.0) : 
		   aStat(name, false)
    {      
      aName= new char[strlen(name)+1];
      strcpy(aName, name);
      
      // printf("creating new nucleus %s\n", aName);
      aSave= save;
      aN_neurons= n_neurons;
      aSlope= slope;
      aThreshold= threshold;
      aSalience_weight= 0.0;
      aTotal_output= 0.0;
      
      for (int i=0; i<aN_neurons; i++) 
	{
	  leaky* l;
	  switch(dt)
	    {
	    case eNone: 
	      l= new leaky; 	  
	      l->init(time_step, slope, threshold);
	      break;
	    case eD1:
	    case eD2: 
	      l= new da_leaky(dt);
	      l->init(time_step, slope, threshold);
	      break;
	    case eD1f:
	    case eD2f: 
	      l= new da_full_leaky(dt);
	      l->init(time_step, slope, threshold, da_gain, da_pivot);
	      break;
	    default: cerr << "unrecognized Da_type in nucleus::add_leaky\n";
	    }
	  aNeurons.push_back(l);
	}
      // cout << "end of Nucleus constructor for " << aName << endl;
     }
  
  inline float nucleus::update_activation(float salience[])
    {
      float d;
      
      aMaxd= 0;
      for (int i=0; i<aN_neurons; ++i)
	{
	  d= aNeurons[i]->integrate(net_input(i, salience[i]));
	  if (d>aMaxd) aMaxd= d;
	}
      return aMaxd;
    }
  
  inline void nucleus::update_output()
    {
      float v;
      aTotal_output= 0.0;
      for (int i=0; i<aN_neurons; ++i) 
	aTotal_output+= v= aNeurons[i]->piecewlin();
    }
  
  inline void nucleus::update_stat()
    {
      for (int i=0; i<aN_neurons; ++i) aStat.add_value(aNeurons[i]->output());
    }

  inline void nucleus::set_thresholds(float th)
    {
      for (int i=0; i<aN_neurons; ++i) aNeurons[i]->set_threshold(th);
    }

  inline float* nucleus::values(Leaky_value lv, float values[])
    {
      int i;
      for (i=0; i<aN_neurons; ++i) values[i]= aNeurons[i]->value(lv);
      return values;
    }
  
  inline float* nucleus::invert_output(float values[], float thresh)
    // *** changed 27.01.03 to use thresh parameter
    // thresh defaults to -99
    // used for converting GPi outputs into a motor gating signal
    {
      int i;
      for (i=0; i<aN_neurons; ++i) values[i]= aNeurons[i]->invert_output(thresh);
      return values;
    }
  
  inline float* nucleus::threshold_output(float values[], float offset)  
    // offset defaults to 0
    // used for converting THL outputs into behavioural feedback signals
    {
      for (int i=0; i<aN_neurons; ++i) values[i]= aNeurons[i]->threshold_output(offset);
      return values;
    }     
  
  inline float* nucleus::pwl_values(float values[])
    // transform an array of values using the PWL function for this nucleus
    {
      for (int i=0; i<aN_neurons; ++i) values[i]= aNeurons[0]->piecewlin(values[i]); 
      return values;
    }
  

  // monitoring and output
  
  void nucleus::print(int verbose= 1)
    {
      if (verbose==1) cout << name();
      else if (verbose>1)
	{
	  cout << "nucleus: " << name() << " " << aN_neurons << " neurons(";
	  aNeurons[0]->print();
	  if (verbose==2) 
	    {
	      cout << "), inputs(";
	      if (aSalience_weight!=0) cout << "salience, ";
	      pprint_vector(aInputs, "", ", ", ") ");
	      cout << endl;
	    }
	  else
	    { 
	      cout << ")\n";
	      if (aSalience_weight!=0)
			cout << "  input: salience(" << aSalience_weight << ")\n";
	      print_vector(aInputs, 2);
	    }
	}
    }
  
  void nucleus::print_values(Leaky_value lv = eY)
    {
      int i;
      
      cout << name() << ": ";
      for (i=0; i<aN_neurons; i++) printf("%7.4f ", aNeurons[i]->value(lv));
      printf("\n");
    }


}   // end of namespace

#endif

















