/*****************************************************************************
 * File:         controller.h                                                *
 * Version:      3.0                                                         *
 * Description:  class for controlling robot bg experiments                  *
 * Authors:      Tony Prescott & Fernando Montes                             *
 * Contact:      t.j.prescott@shef.ac.uk                                     *
 *****************************************************************************/


#ifndef CONTROLLER
#define CONTROLLER

// see build.cpp for the customisable method for adding subsystems, behaviours, etc.

#include <Khepera.h>
#include <KheperaGripper.h>
#include <vector>
#include <string>
#include <list>

#include "controller_gui.h"
#include "experiment.h"        // base class for recording experiments
#include "../utils/bout.h"     // *** 27.01.03 class for recording behavioural bouts

#define MAX_ARR_SZ 30
#define WRITE_MOTORS OFF
#define ROBOT_STEP 8           // parameter passed to Khepera_step

using namespace bgs;
using namespace std;

typedef list <bout*> btlist;  // *** 27.01.03 new type for a list of bouts 

class controller : public experiment, data_log
{
 private:
  // Private attributes 
  string aName;                            // identifying attributes
  string aVersion;
  string aDate;
  string aDescription;
  
  // pointers to objects 
  observer_log*  aObserver;                // ptr to Observer file handler 
  sensory_sys*   aSenses;                  // ptr raw sensory system 
  pvector        aPercepts;                // vector of ptrs perceptual subsystems 
  mvector        aMotivations;             // vector of ptrs to motivational subsystems 
  svector        aSubsystems;              // vector of ptrs to perceptual & motivational subsystems 

  subsys         *aMonsys;                 // used to identify a subsys for monitoring
  bvector        aBehaviours;              // vector of ptrs to behavioural subsystems 
  basal_ganglia* aBg;                      // ptr to basal ganglia object 
  motor_plant*   aMplant;                  // ptr to motor plant object

  bool   aWta;                             // set to TRUE for hard switching (default soft switching) 

  float          aSalience_scale;          // scale all salience signals
  float          aSalience_inc;            // increment all salience signals

  // *** 27.01.03 removed aUse_persistence attribute
  float*         aPersistence;             // array of persistence weights
  float*         aRaw_salience;            // array of salience values without the busy signal
  float*         aSalience;                // array of full salience values
  float*         aGates;                   // array of behaviour gating signals
  float*         aFeedback;                // array of behavioural feedback signals
  int            aFeedback_source;         // *** new flag to determine whether feedback is 
                                           // computer using thalamic or GPi activity

  // analysis attributes 
  int aDisengaged_steps;                   // steps with no engaged (thalamic feedback) behaviours
  int aMult_engaged_steps;                 // steps with multiple engaged behaviours
  int aLast_winner;                        // winning behaviour on previous time-step 
  int aN_switches;                         // number of behavioural switches during the experiment 
  int aN_bouts;                            // number of behavioural bouts during the experiment 
  double aActivity_secs;                   // seconds with one or more active behaviours
  float  aNull_gate;                       // activity of the NULL behaviour;

  // timers and statistics
  timer    aElapsed;                       // timing periods of activity between movement on/off
  timer_stat aRobot_timer;                 // timing and statistic for khepera_step calls

  btlist aBouts;                           // *** 27.01.03 list of behavioural bouts 

  // Private methods 
  void set_name(string nm, string vs, string dat) { aName= nm; aVersion= vs; aDate= dat; }
  void set_description(string ds) { aDescription= ds; }
  
  // *** new
  int set_feedback_source(int f) { return aFeedback_source= f; }
  
  basal_ganglia* add_basal_ganglia(basal_ganglia* bg) { return aBg= bg; }
  subsys* add_subsystem(subsys* sptr) { aSubsystems.push_back(sptr); return sptr; }
  percept* add_percept(percept* pptr) 
    { aSubsystems.push_back(pptr); aPercepts.push_back(pptr); return pptr; }
  motivation* add_motivation(motivation* mptr) 
    { aSubsystems.push_back(mptr); aMotivations.push_back(mptr); return mptr; }
  behaviour* add_behaviour(behaviour* bptr) { aBehaviours.push_back(bptr); return bptr; }

  // *** new ADD_BOUT method
  bout* add_bout(double et, int n, behaviour* bv, int wait) 
    { 
      bout* bptr= new bout(et, n, bv->label(), bv->id(), wait); 
      aBouts.push_back(bptr); return bptr; 
    }

  float* persistence() { return aPersistence; }

  bool set_persistence();
  void initialise_arrays();
  void update_sensory_systems();

  void reset_behaviours();
  bool check_movement();
  void update_behaviours();

  void update_basal_ganglia();   

  void boot_robot();
  void init_robot();
  void set_robot_signals(int);
  void update_robot();

  void update();                          // main update method called on each controller step 

  void update_analysis();

  void print_description(int);
  void print_behaviours(int); 
  void print_subsystems(int); 
  void print_behaviour_weights(int);

  void start_analysis();
  int analyse_behaviours(double);
  void end_analysis();
  void print_analysis_hdr(bool, int);
  void print_analysis_r(bool, int);
  void print_bouts(int, string, bool);
  void print_transitions(string, bool, bool);
  void print_surf(int, string, bool);
  void print_analysis(int);
  virtual void log_header();
  void log(double);

  void build();                           // defined in CTRL_BUILDS/***.H (where *** indicates variant)
  
  virtual void experiment_restart();
  virtual void experiment_start();
  virtual void experiment_step();
  virtual void experiment_end(bool);

 public:
  bool winner_takes_all() { return aWta; }
  bool set_winner_takes_all(bool b) { return aWta= b; }
  float salience_scale() { return aSalience_scale; }
  float set_salience_scale(float f) { return aSalience_scale= f; }
  float salience_inc() { return aSalience_inc; }
  float set_salience_inc(float f) { return aSalience_inc= f; }
  void print(int);
  void print(int, string, bool);
  void print_stats_hdr(int);
  void print_stats_r(int);
  void print_stats(bool, int);
  void print_stats(bool, int, string, bool);
  string name() { return aName+"_"+aVersion; }
  string efolder() { string ef= aInterface->efolder_name(); return (ef.empty())? name() : ef; }
  void start(interface*, observer_log*);
  int n_behaviours() { return aBehaviours.size(); }
  void run() { for(;;) update(); }

  // constructor
  controller() :
    aSenses(new sensory_sys), aMplant(new motor_plant), aFeedback_source(1), 
    aElapsed("Elapsed time s"), aRobot_timer("Robot time ms", 3) 
    { aSalience_scale= 1.0; aSalience_inc= 0.0; aWta= false; aMonsys= NULL; }
};

bool controller::set_persistence()
{
  float p;
  bool flag= false;
  for (unsigned i=0; i<aBehaviours.size(); i++) 
    {
      p= aPersistence[i]= aBehaviours[i]->persistence();
      if (p!=0.0) flag= true;
    }
  if (!flag) { delete[] aPersistence; aPersistence= NULL; return false; }
  else return true;
}

void controller::initialise_arrays()
{
  // can only be called after build is complete
  int noc= aBehaviours.size();
  
  if (noc==0) cout << "controller warning: initialising arrays before behaviours\n";
  aSalience= new float[noc];
  aRaw_salience= new float[noc];
  aGates= new float[noc];
  aFeedback= new float[noc];
  aPersistence= new float[noc];
  set_persistence();
}

// START: CALLS BUILD AND PERFORMS FURTHER INITIALISATION

void controller::start(interface* iface, observer_log* obs)
{
  aInterface= iface;
  aObserver= obs;

  /* Call the customised build method */ 
  build();
  boot_robot();

  printf("\nRobot ready...\n");

  // NOTE: THE PROGRAM WILL PAUSE HERE UNTIL YOU CLICK GO IN THE WORLD WINDOW 
  khepera_step(ROBOT_STEP);

  // create the user interface
  aInterface->make_windows(name(), aFlags, aPercepts, aBehaviours, aBg);

  // initialise parameters 

  // cout << "updating senses\n";
  aSenses->update();
  
  // cout << "updating percepts\n";
  for (unsigned i=0; i<aPercepts.size(); i++) aPercepts[i]->update();

  // cout << "resetting the behaviours";
  reset_behaviours();

  aElapsed.start();
}

void controller::update_sensory_systems()
  // update basic sensory system and  perceptual and motivational sub-systems 
{
  aSenses->update();
  for (unsigned i=0; i<aSubsystems.size(); i++) 
    {
      aSubsystems[i]->update();
      // if (talk()) cout << aSubsystems[i]->value() << " ";
    }
  // cout << endl;
  aInterface->update_percepts_window(aPercepts);
  if (completed() && movement()) 
    aInterface->update_graph_window(aMotivations); // ***   
}

void controller::reset_behaviours()
{
  for (unsigned i=0; i<aBehaviours.size(); i++) aBehaviours[i]->reset();
  // *** 27.01.03 added motivation reset
  for (unsigned i=0; i<aMotivations.size(); i++) aMotivations[i]->reset();
  if (aMonsys!=NULL) aMonsys->zero();
}

bool movementtm1= false;   // global variable used to check if the movement flag has changed

bool controller::check_movement()
{
  // if movement has been enabled then reset the behaviours
  if (movement() && !movementtm1) 
    {
      if (completed()) init_robot();
      reset_behaviours();
      // aInterface->remake_windows(name(), aFlags, aPercepts, aBehaviours, aBg);
      aElapsed.start();
    }
  else aElapsed.stop();
  return movementtm1= movement();
}

void controller::update_behaviours()
{
  const int verbose=0;

  if (verbose) cout << "update motor commands based on current sense data\n";
  for (unsigned i= 0; i<aBehaviours.size(); i++) aBehaviours[i]->update_motor(aSenses);
  
  if (verbose) cout << "gate motor outputs and calculate the new motor vector\n"; 
  aMplant->reset();
  aBg->gates(aGates);  // *** can pass talk() as second parameter

  int winner= aBg->winner();

  for (unsigned i= 0; i<aBehaviours.size(); i++) 
    {
      float g= aGates[i];
      if (aWta) { (((int) i)==winner)? g= 1.0 : g= 0.0; }    // enforce hard-switching if required
      aBehaviours[i]->gate_output(g, aMplant->motors());
    }

  if (verbose) cout << "provide behaviours with (thalamic) feedback\n"; 
  aBg->feedback(aFeedback);
  for (unsigned i= 0; i<aBehaviours.size(); i++)
    if (aFeedback_source==1)
      aBehaviours[i]->feedback(aFeedback[i]);
    else
      aBehaviours[i]->feedback(aGates[i]);            
    // *** 27.01.03 possible to use the GPi to trigger feedback in Vanilla model

  if (verbose) cout << "update the motor plant\n"; 
  aMplant->step(movement());

  if (verbose) cout << "update display\n"; 
  if (WRITE_MOTORS==ON) aMplant->print();
  if (talk()) print_vector(aBehaviours, 0);
  if (movement()) aInterface->update_ethogram_window(aBehaviours); // ***
}


// UPDATING THE BASAL GANGLIA MODEL

void controller::update_basal_ganglia()
{
  const int verbose= 0;
 
  float ss= salience_scale(), si= salience_inc();

  if (verbose) cout << "Update salience\n";
  
  for (unsigned i= 0; i<aBehaviours.size(); i++)
    {
      aSalience[i]= ss * (si + aBehaviours[i]->update_salience());
      aRaw_salience[i]= ss * (si + aBehaviours[i]->raw_salience());
      // (talk()) cout << aSalience[i] << "(" << aRaw_salience[i] << ") ";
    }
  if (verbose) cout << "Converge bg\n";
  aBg->converge(aSalience);

  if (verbose) cout << "Update bg interface\n";
  aBg->pwl_salience(aRaw_salience);
  aBg->pwl_salience(aSalience);
  if (verbose) cout << "Update salience_window\n";
  aInterface->update_salience_window(aBehaviours, aRaw_salience, aSalience);
  if (verbose) cout << "Update monitor_window\n";
  aInterface->update_monitor_window();
}


// INTERFACING WITH THE ROBOT

void controller::boot_robot() 
  // note that for the real robot sensors are enabled anyway 
{
  khepera_live();                              // activate the khepera 
  khepera_enable_proximity(IR_ALL_SENSORS);    // enable proximity measures 
  khepera_enable_light(IR_ALL_SENSORS);        // enable light measures 
  gripper_enable_presence();                   // enable gripper 
  gripper_enable_resistivity();                // enable gripper 
}

// *** new 18.03.03

void controller::init_robot()
{
  aMplant->initialise();
  khepera_step(ROBOT_STEP);
  aMplant->reset();
}

void controller::set_robot_signals(int winner)  
     /* sets the LEDS on the Khepera according to the number of the winner */
{
  khepera_set_led(LEFT, (int) winner % 2);
  khepera_set_led(RIGHT, (int) winner / 2);
}

void controller::update_robot()
{
  aRobot_timer.start();
  khepera_step(ROBOT_STEP);
  aRobot_timer.stop();
}


// STEP: THE MAIN UPDATE METHOD

int it= -1;

void controller::update()
{
  const int verbose= 0;
  if (verbose) cout << "Update sensory systems\n";
  update_sensory_systems();  // update raw senses/percept/motivation systems

  if (verbose) cout << "Update basal ganglia\n";
  update_basal_ganglia();    // compute saliences then
                             // run basal ganglia till convergence 

  if (verbose) cout << "Update behaviours\n";
  update_behaviours();       // generate motor commands for all behaviours, 
                             // gate behaviour output with BG output,  
			     // compute aggregate command, update motorplant 
  if (verbose) cout << "Update robot\n";
  update_robot();            // run the robot 
  check_movement();

  if (verbose) cout << "Update experiment\n";
  if (estart()) experiment_restart();
  if (completed() && movement())
    {
      // int et= (int) aElapsed.check();
      // if (et != it) { cout << "time " << (int) aElapsed.check() << endl; it= et; }
      show_etime(aMovement, aElapsed.check(), 0, 0, "time");
      aInterface->update_time_bar(aElapsed.check());
    }
  experiment_update();      // run experiment methods 
}


// MONITORING

void controller::print_description(int verbose= 1)
{
  // cout << endl << date_and_time_str() << endl;
  cout << endl << date_str() << '\t' << time_str() << endl;
  if (verbose==0)
    {
      cout << "\nController:    " << aName << " " << aVersion << ", updated " << aDate << endl;
      cout << "               " << aDescription << "\n";
    }
  else {
    cout << "\nBrain Inspired Robot Action Selection (BIRAS)\n";
    cout << "Version: " << aName << " " << aVersion << ", updated " << aDate << "\n";
    cout << "Description: " << aDescription << "\n";
    if (aWta) cout << "Hard switching version (winner-takes-all)\n";
  }
}

void controller::print_behaviours(int verbose= 1)
{ 
  if (verbose>=1)
    {
      cout << "Behaviours(" << aBehaviours.size() << "): "; 
      pprint_vector(aBehaviours); 
    }
}

void controller::print_subsystems(int verbose= 1) 
{ 
  if (verbose>=1)
    {
      cout << "Percepts(" << aPercepts.size() << "): ";
      pprint_vector(aPercepts);
      if (!aMotivations.empty()) {
	cout << "Motivations(" << aMotivations.size() << "): ";
	pprint_vector(aMotivations);
      }
    }
}

void controller::print_behaviour_weights(int verbose= 1)
{
  if (verbose>=1) {
    cout << "\nInputs and weights:\n";
    print_vector(aBehaviours, 2);
    cout << "\n";
  }
}

void controller::print(int verbose= 1)
{
  // printf("verbosity is %d\n\n", verbose);
  print_description(verbose);
  print_behaviours(verbose);
  print_subsystems(verbose);
  print_behaviour_weights(verbose - 1);
  aBg->print(verbose);
}

void controller::print(int verbose, string fname, bool append= false)
{
  ofstream outs;
  if (redirect_cout(outs, fname, append)) {
    print(verbose);
    reinstate_cout(outs);
  }
}

// ANALYSIS METHODS

void controller::start_analysis()
{
  aDisengaged_steps= aMult_engaged_steps= 0;                   // ***
  aN_switches= aN_bouts= 0;
  aActivity_secs= 0.0;
  aLast_winner= -1;
  if (!aBouts.empty())       // ***
    {
      // not totally sure that this first bit is necessary, but what the hell...
      for (btlist::iterator pos= aBouts.begin(); pos != aBouts.end(); ++pos) 
	{ (*pos)->~bout(); (*pos)= NULL; }
      aBouts.clear();
    }
}

int wait_steps= 0;

int controller::analyse_behaviours(double et)
     // et is experiment time in seconds
     // *** major changes to include bout information
{

  float eff= aBg->efficiency();
  float dis= aBg->distortion();
  int winner= aBg->winner();
  int flag= aBg->sflag();

  aNull_gate= (eff<1.0)? 1.0 - eff : 0.0;      // activity of the NULL behaviour
 
  if (winner != aLast_winner)
    {
      if (!aBouts.empty()) 
	aActivity_secs += aBouts.back()->finish();
      aN_switches+= 1;
      aLast_winner= winner;
      if (winner != -1) 
	{
	  aN_bouts+= 1;
	  if (aBouts.empty()) wait_steps= 0;
	  add_bout(et, aN_bouts, aBehaviours[winner], wait_steps);
	  wait_steps= 0;
	}
    }
  
  if (flag > 0)                                                 // something selected
    {
      bout *bt= NULL;                                        
      if (aBouts.empty()) cout << "Warning: empty bout list\n";
      else bt= aBouts.back();
      if (bt!=NULL) bt->update(et, eff, dis);
      wait_steps= 0;
    }
  else wait_steps++;                                            // count steps with no winner
  
  if (flag >= 4 && talk())                                      // multiple selection
    {
      printf("Multiple winners:\n");
      print_vector(aBehaviours, 0);
    }
  
  // *** added 27.01.03 but probably redundant
  int n_engaged= 0;
  for (unsigned i=0; i<aBehaviours.size(); i++) if (aFeedback[i]>0) n_engaged+= 1;
  if (n_engaged == 0) aDisengaged_steps+= 1;
  else if (n_engaged > 1) aMult_engaged_steps+= 1;
  
  set_robot_signals(winner);
  return winner;
}

void controller::end_analysis()
{
  if (!aBouts.empty() && !aBouts.back()->completed()) 
    aActivity_secs += aBouts.back()->duration();    // ***
}

void controller::print_analysis_hdr(bool el= false, int verbose= 0)
{
  cout << "dur(s)\tdur(i)\tact(s)\tsw(n)\tbt(n)\tmnbt(s)";
  if (verbose>1) cout << "\tdiseg%\tmlteg%";
  if (el) cout << endl; else cout << '\t';
}

void controller::print_analysis_r(bool el= false, int verbose= 0)
{
  // int acts= aBg->activity_steps();
  int tsteps= experiment_end_steps();

  cout << experiment_end_time() << '\t' << tsteps << '\t';
  cout << aActivity_secs << '\t';
  cout << aN_switches-1 << '\t' << aN_bouts << '\t';
  cout << aActivity_secs/aN_bouts;
  if (verbose>1) cout << '\t' << ((float) aDisengaged_steps/tsteps) << '\t';
  if (verbose>1) cout << ((float) aMult_engaged_steps/tsteps);
  if (el) cout << endl; else cout << '\t';
}

void controller::print_transitions(string fname, bool append= false, bool labels= false)
{
  ofstream outs;
  int trans[10][10];
  int nb= aBehaviours.size() + 1;
  int max_gap= 3;                      // maximum gap before a bout of null behavior is noted
  int wait, curr, last= -1;

  if (nb>10) cerr << "Warning to many behaviours for transition table\n";
  else
    {
      // cerr << "\nbuild transition table\n";
      for (int i=0; i<nb; i++) for (int j=0; j<nb; j++) trans[i][j]= 0;
      for (btlist::iterator pos= aBouts.begin(); pos != aBouts.end(); ++pos)
	{
	  curr= (*pos)->id(); wait= (*pos)->wait();
	  if (last>=0 && last<10 && curr>=0 && curr<10) 
	    {
	      if (wait<=max_gap) trans[last][curr]++;
	      else { trans[last][0]++; trans[0][curr]++; }
	    }
	  last= curr;
	}
      
      // cerr << "write transition table to file " << fname << endl;
      if (redirect_cout(outs, fname, append)) {
	if (labels) 
	  {
	    cout << "\n*\tNull";
	    for (unsigned i=0; i<aBehaviours.size(); i++) 
	      cout << "\t" << aBehaviours[i]->label();
	    cout << endl;
	  }
	table_format_flags(3, false);
	for (int i= 0; i<nb; i++)
	  {
	    if (labels) { if (i==0) cout << "Null\t"; else cout << aBehaviours[i-1]->label() << '\t'; }
	    for (int j=0; j<(nb - 1); j++) cout << trans[i][j] << '\t';
	    cout << trans[i][nb - 1] << endl;
	  }
	reset_format_flags();
	reinstate_cout(outs);
      }
      // cerr << "finished writing transitions\n";
    }
}

void controller::print_bouts(int verbose, string fname, bool append= false)
{
  ofstream outs;
  if (redirect_cout(outs, fname, append)) {
    cout << "bn\tstrt(s)\tend(s)\tbhr\tid\tdur(s)\tdur(i)\twt(i)";
    cout << "\tE_mn\tE_md\tD_mn\tD_md\tNotes\n";
    table_format_flags(3, false);
    for (btlist::iterator pos= aBouts.begin(); pos != aBouts.end(); ++pos) 
      (*pos)->print();
    reset_format_flags();
    reinstate_cout(outs);
  }
}

void controller::print_surf(int verbose, string fname, bool append= false)
{
  ofstream outs;
  if (redirect_cout(outs, fname, append))
    {
      bool normalised= false;
      bool labels= false;
      // for SigmaPlot include labels, leave out for MatLab
      table_format_flags(3, false);
      aBg->surface()->print(normalised, labels);
      if (verbose>0)
	{
	  cout << endl;
	  normalised=~normalised;
	  aBg->surface()->print(normalised, labels);
	}
      reset_format_flags();
      reinstate_cout(outs);
    }
}

void controller::print_analysis(int verbose)
{
  int acts= aBg->activity_steps();
  int tsteps= experiment_end_steps();

  cout << "\n";
  cout << "total_time(secs)  " << tsteps << endl;
  cout << "total_steps       " << experiment_end_steps() << endl;
  cout << "active_time(secs) " << aActivity_secs << endl;
  cout << "active_steps      " << acts << endl;
  cout << "total_switches    " << aN_switches << endl;
  cout << "total_bouts       " << aN_bouts << endl;
  cout << "mean_bout(secs)   " << aActivity_secs/aN_bouts << "\n";
  cout << "mean_bout(steps)  " << ((float) acts/aN_bouts) << "\n";
  if (verbose>0)
  {
    cout << "disengaged%       " << ((float) aDisengaged_steps/tsteps) << endl;     // ***
    cout << "mult_engaged%     " << ((float) aMult_engaged_steps/tsteps) << "\n\n";
  }
}

void controller::print_stats_hdr(int verbose= 0)
{
  // cerr << "printing stats_hdr\n";
  cout << "date\ttime\t";
  print_analysis_hdr(false, verbose);
  if (aMonsys!=NULL) aMonsys->print_analysis_hdr(false, verbose);
  for (unsigned i=0; i<aMotivations.size(); i++) cout << aMotivations[i]->name() << "\t";
  aBg->print_analysis_hdr(true, verbose);
}

void controller::print_stats_r(int verbose= 0)
{
  // cerr << "printing stats\n";
  cout << date_str() << '\t' << time_str() << '\t';
  table_format_flags(3, false);
  print_analysis_r(false, verbose);
  if (aMonsys!=NULL) aMonsys->print_analysis(false, verbose);
  for (unsigned i=0; i<aMotivations.size(); i++) cout << aMotivations[i]->mean() << "\t";
  aBg->print_analysis_r(true, verbose);
  reset_format_flags();
}

void controller::print_stats(bool full_time, int verbose= 1)
{
  if (verbose>0) cout << endl << date_and_time_str() << endl;
  table_format_flags(3, false);
  experiment::print(full_time, verbose);
  if (verbose>0)
    {
      aRobot_timer.print(verbose);
      print_analysis(verbose);
      aBg->print_analysis(verbose);
      cout << endl;
    }
  reset_format_flags();
}

void controller::print_stats(bool full_time, int verbose, string fname, bool append= true)
{
  ofstream outs;
  // cerr << "print_stats\n";
  bool newf= (!file_exists(fname) || !append);
  if (redirect_cout(outs, fname, append)) {
    if (newf)
      print_stats_hdr(verbose);
    print_stats_r(verbose);
    reinstate_cout(outs);
  }
}

void controller::log_header()
{
  ofstream& outs= log_stream();
  if (record_log() && outs)
    {
      outs << "t" << '\t' << "win";
      outs << '\t' << "eff" << '\t' << "dis" << '\t';
      for (unsigned i=0; i<aBehaviours.size(); i++) 
	outs << aBehaviours[i]->name() << '\t';  
      outs << "null\t";
      for (unsigned i=0; i<aPercepts.size(); i++) 
	outs << aPercepts[i]->name() << '\t';  
      if (!aMotivations.empty()) 
	for (unsigned i=0; i<aMotivations.size(); i++) 
	  outs << aMotivations[i]->name() << '\t';
      outs << "wS_\tnS_\t";
     outs << endl;
    }
}

void controller::log(double et)
{
  ofstream& outs= log_stream();
  if (record_log() && outs) 
    {
      outs << et << '\t' << aLast_winner;
      outs << '\t' << aBg->efficiency();
      outs << '\t' << aBg->distortion() << '\t';
      osprint(outs, aBehaviours.size(), aGates, '\t');
      outs << aNull_gate << '\t';
      outs.setf(std::ios::showpos);
      for (unsigned i=0; i<aPercepts.size(); i++) 
	outs << (int) aPercepts[i]->value() << '\t';  
      outs.unsetf(ios::showpos);
      if (!aMotivations.empty()) 
	for (unsigned i=0; i<aMotivations.size(); i++) 
	  outs << aMotivations[i]->value() << '\t';
      outs << aBg->win_sal() << '\t' << aBg->next_sal();
      outs << endl;
  }
}

// Methods for running the experiment

void controller::experiment_restart()
{
  experiment::experiment_restart();
  init_robot();
  aInterface->init_graph_window();
  aInterface->init_ethogram_window(aBehaviours);
}

void controller::experiment_start()
{
  experiment::experiment_start();
  
  // aInterface->set_button_enabled(aEstart.id(), false);
  start_analysis();
  aBg->start_log();
  start_log();
  aObserver->start(aBehaviours); 
  aRobot_timer.zero();
  set_movement(true);
  aInterface->set_button_enabled(aMovement.id(), false);
  aInterface->update_button(aMovement);
  aInterface->init_time_bar(experiment_duration());
  reset_behaviours();
  aBg->zero();
 }

void controller::experiment_step()
{
  experiment::experiment_step();

  double et= experiment_time();
  int winner= analyse_behaviours(et);
  aInterface->update_time_bar(et, winner);
  aInterface->update_graph_window(aMotivations, et); // ***   
  aBg->log(et);
  aObserver->step(aBehaviours, et);
  log(et);
}

void controller::experiment_end(bool full_time)
{
  experiment::experiment_end(full_time);
  end_analysis();

  set_movement(false);
  aInterface->set_button_enabled(aMovement.id(), true);
  aInterface->update_button(aMovement);
  aInterface->set_button_enabled(aEstart.id(), true);
  aInterface->update_button(aEstart);

  aObserver->finish(experiment_end_time());
  bool comment= false;
  if (comment) cout << "\ncompleting basal ganglia log\n";
  aBg->finish_log();
  finish_log();
  if (comment) cout << "\nsaving bouts to /bouts.txt\n";
  print_bouts(0, efolder()+"/bouts.txt");
  if (comment) cout << "\nsaving transitions to /transitions.txt\n";
  print_transitions(efolder()+"/transitions.txt", false, true);
  if (comment) cout << "\nsaving surface to /surf.txt\n";
  print_surf(0, efolder()+"/surf.txt");
  if (comment) cout << "saving model to /model.txt\n";
  print(3, efolder()+"/model.txt");
  if (comment) cout << "\nsaving results to results.txt\n";
  print_stats(full_time, 0);
  print_stats(full_time, 0, efolder()+"/result.txt", false);
  if (full_time) print_stats(full_time, 0, "results.txt", true);
 cout << "data in folder '" << efolder() << "'.\n";
}


#endif




















