/*****************************************************************************
 * File:         ctrl/ctrl_motivated.h                                       *
 * Version:      4.1                                                         *
 * Description:  Controller method for customising model                     *
 * Authors:      Tony Prescott                                               *
 * Contact:      t.j.prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef CTRL_BASIC
#define CTRL_BASIC

void controller::build()
{
  const int mode= 3;
  const char *modes[4] = {"unmotivated", "fearful", "hungry", "fearful and hungry"};

  // set the controller name etc.
  set_name("Motivated", "5.0", "Date");

  set_feedback_source(1);              // 0 - GPi, 1 - thalamus
  set_salience_scale(1.0);             // scale all salience values by this amount
  set_salience_inc(0.0);               // increment all salience values by this amount
  set_winner_takes_all(false);         // true if hard-switching enforced (use false normally)

  // note: computation is ss * (si + sal)
  string fs= (aFeedback_source==1)? "thalamic feedback" : "GPi feedback";
  string sss= "", ssi= "";
  if (salience_scale()!=1.0) sss= " + s_scale (" + ftos(salience_scale(), 2) + ")";
  if (salience_inc()!=0.0) ssi=  " + s_inc (" + ftos(salience_inc(), 2) + ")";
  set_description("Khepera foraging behaviour, " + fs + sss + ssi);

  // set the duration of the experiment in seconds
  experiment_set_duration((mode==0)? 40 : 120);   // should be 60, 120 or 300 secs
  set_record_log(true);
  set_record_file(efolder()+"/ctrl_log.txt");

  // add perceptual sub-systems 
  percept* pwall(add_percept(new wall_ps(aSenses)));
  percept* pgrip(add_percept(new grip_ps(aSenses)));
  percept* pcan(add_percept(new can_ps(aSenses)));
  percept* pcorner(add_percept(new corner_ps(aSenses)));
  percept* plight(add_percept(new light_ps(aSenses)));
  
  percept* pnest(add_percept(new nest_ps(plight, pcorner, pwall)));
		  
  
  // add motivational sub-systems

  // following parameters are selected using the value of MODE
  const bool fear_dynamic[4]= { false, true, false, true };
  const bool hunger_dynamic[4]= { false, false, true, true };
  const float fear_init[]= { 0.5, 1.0, 0.0, 1.0 };
  const float hunger_init[]= { 0.5, 0.0, 0.0, 0.2 };

  // fear_monotonic: decreases additively with time
  motivation* mfear(add_motivation(new motivation("Fear")));
  mfear->set_range(0.0, 1.0);                           // minimum and maximum values
  mfear->add_mood(1.0, -0.00070);                       // default mood decreasing linearly, .00075
  mfear->set_dynamic(fear_dynamic[mode]);               // true if time-varying
  mfear->init_value(fear_init[mode]); 
  
  // hunger: increases multiplicatively with time
  motivation* mhunger(add_motivation(new hunger_ms(pgrip, pnest)));
  mhunger->set_range(0.0, 1.0);                         // minimum and maximum values
  mhunger->add_mood(1.0, +0.0015);                      // default mood increasing linearly, 0.002
  mhunger->add_mood(1.0, -1.0);                         // on can deposit add -0.5
  mhunger->set_dynamic(hunger_dynamic[mode]);           // true if time-varying
  mhunger->init_value(hunger_init[mode]); 

  // add behavioural sub-systems with appropriate perceptual/motivational inputs and weights

  // can seek
  behaviour* bcans(add_behaviour(new can_seek(1)));
  bcans->add_input(pcan,    -0.12);       // can SHOULD NOT be visible  
  bcans->add_input(pgrip,   -0.12);       // gripper should be EMPTY         
  bcans->add_input(mfear, -0.06);         // should not be fearful
  bcans->add_input(mhunger, +0.45);       // should be hungry
  bcans->set_threshold(0.0);              // 

  // can pickup                      
  behaviour* bcanp(add_behaviour(new can_pickup(2)));
  bcanp->add_input(pcan,    +0.21);       // can SHOULD be visible
  bcanp->add_input(pgrip,   -0.15);       // gripper should be EMPTY
  bcanp->add_input(mfear, -0.18);
  bcanp->add_input(mhunger, +0.18);       // should be hungry
  bcanp->set_busy_weight(+.78);          // busy signal needed as can becomes invisible, was .78
  bcanp->set_threshold(+0.25);            // 
  // bcanp->set_persistence(+2.23);           // +2.23 for THL_MOT

  // wall seek
  behaviour* bwall(add_behaviour(new wall_seek(3)));
  bwall->add_input(pwall, -0.12);         // wall SHOULD NOT be visible
  bwall->add_input(pgrip, +0.14);         // gripper should be full, was 0.12  
  bwall->add_input(mfear, +0.18);         // salience INCREASES with fear
  bwall->set_threshold(+0.25);            // should be +0.25

  // nest seek (wall follow)
  behaviour* bnest(add_behaviour(new nest_seek(4)));
  bnest->add_input(pwall,   +0.12);       // wall SHOULD be visible
  bnest->add_input(pgrip,   +0.14);       // gripper should be FULL, was .12 
  bnest->add_input(mfear,   +0.21);       // salience increases with fear  
  bnest->set_busy_weight(+0.25);          // busy signal as wall may be briefly lost, 0.25
  bnest->set_threshold(+0.25);            // 

  // can deposit
  behaviour* bcand(add_behaviour(new can_deposit(5)));
  bcand->add_input(pgrip,   +0.33);       // gripper should be FULL +0.55  (+0.45)
  bcand->add_input(pnest,   +0.33);       // nest should be VISIBLE +0.55  (+0.45)
  bcand->add_input(mhunger, +0.18);       // should be hungry
  bcand->set_busy_weight(+0.4);           // busy signal to raise arm after dropping can, .4
  // bcand->set_persistence(+2.23);           // +2.23 for THL_MOT
  bcand->set_threshold(+0.13);            //
                                          // n.b. doesn't complete if drops food outside nest
                                          // needs busy of 1.32 to do that!
                                          // or 0.36 to interrupt if lose sight of nest mid-action
  // finish controller set-up
  initialise_arrays();

  // the basal ganglia
  basal_ganglia* bg(new basal_ganglia (n_behaviours(), persistence()));
  add_basal_ganglia(bg);

  bg->set_log_level(0);                      // log file output 0- basic, 1- full
  bg->set_record_log(true);
  bg->set_record_file(efolder()+"/bg_log.txt");

  // print a summary of the model
  print(0);                               // 0 description only, output, 1 basic, 2 detailed, 3 max

  printf("\nThe robot is %s.\n\n", modes[mode]);
  // cout << "experiment duration is " << experiment_duration() << endl;
}

#endif















