/*****************************************************************************
 * File:         experiment.h                                                *
 * Version:      3.0                                                         *
 * Description:  Controller methods for running/analysing a robot experiment *
 * Authors:      Tony Prescott                                               *
 * Contact:      t.j.prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef EXPERIMENT
#define EXPERIMENT

enum Estage { eNot_started= 0, eIn_progress= 1, eComplete= 2, eAborted= 3};

class experiment : public controller_gui
{
 private:
  double aDuration;                  // experiment duration in seconds (0 = unlimited)
  int aCountdown_secs;               // number of seconds after start before experiment begins
  Estage aStage;                     // stage in the experiment
  int aSteps;                        // number of steps since experiment began

  timer aTimer;                      // timing the experiment 
  timer_stat aStep_timer;            // timing and statistics for Controller step calls 
  timer aCountdown;                  // timing the countdown

  Estage stage() const { return aStage; }
  Estage set_stage(Estage s) { return aStage= s; }

 public:
  bool   completed() { return (aStage==eComplete); }
  double experiment_duration() { return aDuration; }
  double experiment_set_duration(double secs, int cd= 5) 
    { aCountdown_secs= cd; return aDuration= secs; }
  float  experiment_time() { return aTimer.check(); }
  float  experiment_time_left() { return aDuration - aTimer.check(); }
  float  experiment_end_time() { return aTimer.read(); }
  int    experiment_end_steps() { return aSteps+1; }

  virtual void   experiment_restart(); 
  virtual void experiment_start() { aSteps= 0; }
  virtual void experiment_end(bool) {}

  virtual void experiment_step(); 
  void   experiment_update();
  void   print(bool, int);                          // ***

  // constructor
  experiment(double d= 0, int cd= 5) : 
    aDuration(d), aCountdown_secs(cd), aStage(eComplete), aSteps(0),
    aTimer("Total time s"), aStep_timer("Step time ms", 3), aCountdown("Countdown s")
  {}
  virtual ~experiment() {}
};

void experiment::experiment_restart() 
{
  if (aStage==eIn_progress) set_stage(eAborted);
  else
    {
      cout << "\n" << aDuration << "s experiment starting in " << aCountdown_secs << "s";
      cout.flush();
      aStage= eNot_started; aCountdown.start(); aSteps= 0;
    }
}

void experiment::experiment_step() 
{ 
  show_etime(aEstart, aTimer.check(), aDuration, 5, "Time left");
  aSteps++;

  /*
  if (aSteps%100 == 0) 
    {
      cout << "steps " << aSteps; 
      cout << " time " << experiment_time() << "s\n"; 
    }
  */
}
 
void experiment::experiment_update()
{ 
  if (aStage!=eComplete)        
    { 
      if (aStage==eIn_progress || aStage == eAborted) 
	{
	  if ((aDuration != 0 && aTimer.check()>=aDuration) // stop if time up
	      || aStage == eAborted)
 	    {
	      bool full_time= (aStage == eAborted)? false : true;
	      set_stage(eComplete);
	      aStep_timer.stop();                          // stop timer for final step
	      // aStep_timer.timer::print();
	      aTimer.stop();                               // stop total experiment timer
	      experiment_end(full_time);                   // call customised end function
	    } 
	  else                                             // otherwise STEP!!
	    {
	      aStep_timer.stop();                          // stop timer and save value
	      /*if (aStep_timer.read()>200) 
		{
		  cout << "step " << aSteps;
		  aStep_timer.timer::print();
		  } */
	      aStep_timer.start();                         // start timer again from zero
	      experiment_step();                           // call customised step function
	    }
	}

      else if (aStage==eNot_started)
	{
	  if (aCountdown.check()>=aCountdown_secs)        // start the experiment
	    {
	      cout << "\nin progress";
	      cout.flush();
	      set_stage(eIn_progress);
	      aTimer.start();                             // start total timer
	      aStep_timer.zero();                         // initialise the stats for the step timer
	      aStep_timer.start();                        // start timer for first step
	      experiment_start();                         // call customised start function
	    }
	  else 
	    {
	      bool f=show_etime(aEstart, aCountdown.check(), 
				aCountdown_secs, 1, "Countdown", '*');
	      if (f) aInterface->countdown_time_bar(aCountdown.check(), aCountdown_secs);
	    }
	}
    }                              
}

void experiment::print(bool full_time, int verbose= 0)
{
  if (aStage==eNot_started)
    cout << "Experiment has not started\n";
  else if (aStage==eIn_progress)
    cout << "in progress " << aTimer.check() << "s (step " << aSteps << ")\n";
  else
    {
      if (verbose==0)
	{
	  if (full_time) cout << "\ncompleted";
	  else cout << "\nabandoned";
	  cout << " after " << aTimer.read() << "s " << aSteps << " steps, ";
	}
      else
	{
	  if (full_time) cout << "Completed ";
	  else cout << "Unfinished ";
	  cout << (int) aDuration << "s experiment";
	  cout << endl << endl;
	  aStep_timer.print();
	}
    }
}

#endif



















