/*****************************************************************************
 * File:         behaviour.h                                                 *
 * Version:      4.0                                                         *
 * Description:  class definition for behavioural subsystems                 *
 * Author:       Tony Prescott and Fernando Montes                           *
 * Contact:      T.J.Prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef BEHAVIOUR
#define BEHAVIOUR

#include <fstream.h>
#include <vector>
// #include "../utils/bout.h"

class b_input;
typedef vector <b_input*> bin_vector;
class behaviour;
typedef vector <behaviour*> bvector;


// behaviour input (b_input) class:
// consists of a pointer to a subsystem (percept or motivation)
// and the salience weight to be applied to the current value of that subsystem

class b_input 
{
 private:
  subsys* aS;
  float aW;
 public:
  float product() { return aS->value() * aW; }
  void print(int verbose= 0) { 
    if (verbose>=0) {
      string s(aS->name());
      ppad(s, 6, " ");
      cout << aW; 
    }
  }
  b_input(subsys* s, float w) : aS(s), aW(w) {}
};


// behaviour class 

class behaviour 
{
 private:
  enum Motor_els {eLws, eRws, eG, eA};
  int aId;              // unique integer identifier
  int aActive;          // 1 if at least partially disinhibited by BG, 0 otherwise 
  int aActivetm1;       // same but for t-1 
  float aGate;          // basal ganglia gating signal- 0 fully inhibited, 1 fully disinhibited 
  float aTrig;          // thalamic trigger for stm counter 
 
  int aNinps;           // number of input subsystems
  bin_vector aInputs;   // pointers to all the perceptual and motivational inputs and
                        // and their associated weights
  float aBusy_w;        // busy signal weight
  float aThresh_w;      // threshold weight
  float aPersist_w;     // persistence (thalamocortical) weight
  float aRaw_salience;  // weighted sum of inputs
  float aSalience;      // raw salience + weighted busy signal

  float aVector[N_MOTOR_SIG];    // motor vector with only positive values in range 0-1 
  float aGated[N_MOTOR_SIG];     // motor vector following basal ganglia gating 

  int aCounter;                  // step counter  

 protected:             // allows access by more specialised classes 
  timer aTimer;         // behaviour clock use to time expression of fixed action patterns 
  int   aMode;          // current behavioural mode 
  float aBusy;          // busy signal 
  float aMotor[N_MOTORS];        // basic motor vector values -1.0 to +1.0 
  float clip_motor(float);       // clips a motor signal to be in the range -1.0 to 1.0 
  void  get_vector();   // generates new aVector using the current aMotor values 

 public:
  float raw_salience() { return aRaw_salience; }  // weighted sum of inputs
  float salience() { return aSalience; }          // raw salience + weighted busy signal
  float persistence() { return aPersist_w; }      // persistence (thalamocortical) weight
  b_input* add_input(subsys* sptr, float w)       // adds a b_input object to the list aInputs
    { b_input *bi= new b_input(sptr, w); aInputs.push_back(bi); return bi; }
  float set_busy_weight(float w)                  // set weight for busy signal (default 0.0)
    { return aBusy_w= w; }
  float set_threshold(float w)                    // set the threshold weight (default 0.0)
    { return aThresh_w= w; }
  float set_persistence(float w)                  // set the pesistence (thalamocortical) weight (def 0)
    { return aPersist_w= w; }
  float motor_output(int m) { return aGated[m]; } // returns the mth element of the gated motor vector  
  float busy() { return aBusy; }                  // returns the current value of the busy signal
  int active() { return aActive; }                // returns aActive- 1 if part disinhibited, 0 otherwise
  float gate() { return aGate; }                  // returns aGate the bg gating signal, between 0 and 1 
  virtual char* name() { return "behaviour"; }    // returns the name of the behaviour 
  virtual char* label() { return "bv"; }          // returns the label used in GUI windows
  virtual char* mode_name() { return "none"; }    // returns the name of the current behavioural mode
  
  int   id() { return aId; }

  void print(int);                                // print details
  float update_salience();                        // computes salience values
  virtual void update_motor(sensory_sys*);        // computes the motor output for current sensory state 
  void reset();                                   // resets the short-term memory and busy signal
  void feedback(float);                           // increments (or resets) depending on bg feedback 
  void gate_output(float, float*);                // gates the motor output and  sets the aActive flag

  void print_state();                             // write the current mode etc.  iff active is 1 
  void write_observer(ofstream&, double, int);    // logs behaviour start/stop to Observer input file 
  behaviour(int id) : aId(id), aMode(-1)          // Constructor
    { aTimer.start(); }
  virtual ~behaviour() { }
};


// updating the salience
// *** modified 27.01.03 to make initial value of salience= aThresh_w

inline float behaviour::update_salience()
{
   aRaw_salience= aThresh_w;
   for (uint i=0; i<aInputs.size(); i++) aRaw_salience+= aInputs[i]->product();
   aSalience= aRaw_salience + (aBusy * aBusy_w);
   //if (aSalience!= aRaw_salience)
   //  cout << name() << ": salience+ " << (aSalience - aRaw_salience) << "\n";
   return aSalience;
}


// computing the motor vector 

inline float behaviour::clip_motor(float m)
{
  if (m > 1.0) return 1.0;
  else if (m < -1.0) return -1.0;
  else return m;
}

inline void behaviour::get_vector()
{
  for (int i=0; i<(N_MOTORS - 1); i++)
    {
      aVector[2*i]= (aMotor[i]< 0)? -aMotor[i] : 0.0;  
           // wheels back, gripper open movement 
      aVector[2*i+1]= (aMotor[i]> 0)? aMotor[i] : 0.0; 
           // wheels forward, gripper close movement 
    }

  // set the arm position
  int a0= 2*(N_MOTORS - 1);
  aVector[a0]= (aMotor[eA] == -1.0)? 1.0 : 0.0;        // up (-1.0)
  aVector[a0+1]= (aMotor[eA] == 0.5)? 1.0 : 0.0;
  aVector[a0+2]= (aMotor[eA] == 1.0)? 1.0 : 0.0;       // down (1.0)
}


// updating the behaviour 

inline void behaviour::reset()
     // restart the clock, set the busy signal to 0.0 
{
  aCounter= 0;
  aTimer.start();
  aBusy= 0.0;
}

inline void behaviour::feedback(float trig)
     // Uses the (thresholded) thalamic output to determine if the behaviour is selected 
     // if not call Reset.
{
  aTrig= trig;
  if (aTrig<=0) reset();
  else aCounter++;
}

// note this is shadowed by all behaviours

inline void behaviour::update_motor(sensory_sys *senses) 
{
  int i;
  
  aBusy=0.0; aMode= 0;
  
  for (i=0; i<N_MOTORS; i++) aMotor[i]= 0.0;
  get_vector();
}

inline void behaviour::gate_output(float gate, float* aggregate_v)
{
  int i;

  // get the gated motor vector 
  aGate= gate;
  for (i=0; i<N_MOTOR_SIG; i++) 
    {
      aGated[i]= aVector[i]*aGate;
      aggregate_v[i] += aGated[i];
    }

  // set the active flag 
  aActivetm1= aActive;
  aActive= (aGate>0)? 1 : 0;
}

// analysis, display, and debug 

void behaviour::print(int verbose= 1)
{
  if (verbose==0) print_state();
  else if (verbose==1) cout << name();
  else if (verbose>1)
    {
      string s(name());
      ppad(s, 11);
      table_format_flags(2);
      // cout <<  name() << ": ";
      pprint_vector(aInputs, "", ", ", ", ");
      cout << " busy " << aBusy_w << ", ";
      cout << " persist " << aPersist_w << ", ";
      cout << " thresh " << aThresh_w << ".\n";
      reset_format_flags();
    }
}

void behaviour::print_state()
{
  if (aGate>0) 
    printf("%-11s %3.2f %3d: %-14s %s (gate %4.2f) (trig %4.2f)\n", 
	   name(), aTimer.check(), aCounter, mode_name(), 
	   (aBusy>0)? "+" : " ", aGate, aTrig);
}

void behaviour::write_observer(ofstream& os, double t, int override= OFF)
{
  char s[80];
  const int verbose= OFF;

  if (aActive == ON && (aActivetm1 == OFF || override == ON))
    {
      sprintf(s, "%3.3f %s\n", t, name());
      os << s;
      if (verbose) printf("%s: Observer- %s\n", name(), s);
      os.flush();
    }
  else if (aActive == OFF && aActivetm1 == ON)
    {
      sprintf(s, "%3.3f off\n", t);
      os << s;
      if (verbose) printf("%s: Observer- %s\n", name(), s);
      os.flush();
    }
}


#endif

























