/*****************************************************************************
 * File:         forage.h                                                    *
 * Version:      4.0                                                         *
 * Description:  behaviour classes for khepera 'foraging' behaviour          *
 * Author:       Tony Prescott and Fernando Montes                           *
 * Contact:      T.J.Prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef FORAGE
#define FORAGE

#include <stream.h>
#include "behaviour.h"

class can_seek;
class can_pickup;
class wall_seek;
class nest_seek;
class can_deposit;

float sws(float diff)
{
  const float max_diff= 450.0;
  
  if (diff>max_diff) return 1.0;
  else if (diff<30) return 0.07;
  else
    return diff/max_diff;
}


// can seek class 

class can_seek : public behaviour 
{
 public:
  virtual void  update_motor(sensory_sys*);
  virtual char* name() { return "Can_seek"; }
  virtual char* label() { return "Cs"; } 
  virtual char* mode_name();
  can_seek(int id) : behaviour(id) {}
};

void can_seek::update_motor(sensory_sys *senses) 
{
  int sumd, diff, lit, side;
  float lws, rws, ws;

  aBusy= 0.0;

  side= senses->infNearSide();
  sumd= senses->infSum();
  diff= senses->infDiff();
  lit= senses->ambLit();

  // the next bit is needed because the light confuses the IR distance sensor readings 
  if (USE_LIGHT && lit>0 && (sumd > 500))                   // actions in the nest 
    {                                                       
      switch (side) {
      case LEFT:  lws= -0.27; rws= -1.00; break;            // backup, rotating right
      case RIGHT: lws= -1.00; rws= -0.27;                   // backup, rotating left
      }
      aMode = 0;
    }
  
  else if (sumd > 2000 && lit==0)                                      // actions at a wall 
  {
    ws= sws(diff);
    // cout << diff << " " << sws << endl;
    switch (side) {
    case LEFT:  lws =  ws; rws = -ws; aMode = 1; break;      // rotate right
    case RIGHT: lws = -ws; rws =  ws; aMode = 2;             // rotate left
    }
    // else                                                   // backup  
    // { lws  = rws = -0.5; aMode = 3; }
  }

  // actions at a possible can 
  // rotates toward the object hoping its a can 
  // this can cause the robot to get stuck at walls 
  // disabled in lit areas (nests) since this can confuse the sensors 

  else if ((sumd > 1025) && lit==0)                          // was sensors[NEST_DETECTOR]!=1  
    {
      switch (side) {
      case LEFT:  lws  = -0.20; rws =  0.15; aMode = 3; break;
      case RIGHT: lws  =  0.15; rws = -0.20; aMode = 4;
      }
    }

  else                                                       // fast forward 
    { lws= rws= 1.0; aMode = 5; }
 
  aMotor[eLws]= clip_motor(lws);
  aMotor[eRws]= clip_motor(rws);
  aMotor[eA]= (POSTURE)? -1.0 : 0.0;
  aMotor[eG]= 0.0;

  // after 1 second raise the arm and open the gripper 
  // if (tim>=1.0 && tim<=1.2) { aMotor[eA]= -1.0; aMotor[eG]= -1.0; }

  get_vector();
}

char* can_seek::mode_name()
{
  static char *modes[6]= {"Backup in nest", 
			 "Right at wall", 
			 "Left at wall", 
			 "Left at obj", 
			 "Right at obj", 
			 "Fast forward"};

  return modes[aMode];
}



// can pickup class

class can_pickup : public behaviour
{
 public:
  virtual char* mode_name();
  virtual void  update_motor(sensory_sys*);
  virtual char* name() { return "Can_pickup"; }
  virtual char* label() { return "Cp"; }
  can_pickup(int id) : behaviour(id) {}   
};


void can_pickup::update_motor(sensory_sys *senses)
{
  const double stop= 0.3, backup= 1.1, arm_down= 0.4, close_grip= 1.0, arm_up= 0.8;

  float lws, rws, arm, grip;
  aBusy= 1.0;
  double tim= aTimer.check();

  if (tim < stop)                                        // stop
    {
      aBusy=0; aMode=0; lws= 0.1; rws= 0.1; grip= 0.0;
      arm= (POSTURE)? -1.0 : 0.0;
    }

  else if (tim < stop+backup)                            // backup - to make room for arm
    {
      aMode= 1; lws= -0.2; rws=-0.2; grip= -1.0;
      arm= (POSTURE)? -1.0 : 0.0;
    }
  
  else if (tim < stop+backup+arm_down)                   // lower arm
    { 
      aMode= 2; lws= 0.0; rws= 0.0; arm= 1.0; grip= 0.0;
    }
  
  else if (tim < stop+backup+arm_down+close_grip)        // grip the can
    {
      aMode= 3; lws= 0.0; rws= 0.0; grip= 1.0;
      arm= (POSTURE)? 1.0 : 0.0;
      // grip= (senses->GripResist()==1)? 0.0: 1.0;      condition to reduce motor wear - AJP
    }
  
  else if (tim < stop+backup+arm_down+close_grip+arm_up) // arm up  
    {
      aMode= 4; lws= 0.0; rws= 0.0; arm= -1.0; grip= 0.0;
    }
  else                                                   // finished-  set the busy signal to zero 
    {                                                     
      aMode= 5; lws= rws= grip= 0.0;
      arm= (POSTURE)? -1.0 : 0.0;
      aBusy = 0.0;                                       // set the busy signal to 0.0
      reset();
    }

  aMotor[eLws]= lws;
  aMotor[eRws]= rws;
  aMotor[eA]= arm;
  aMotor[eG]= grip;
  get_vector();
}

char* can_pickup::mode_name()
{
  static char *modes[] =  {"Slow approach",
			   "Backup", 
			   "Arm down", 
			   "Close gripper", 
			   "Arm up", 
			   "Finished"};

  return modes[aMode];
}



// wall_seek class

class wall_seek : public behaviour
{
 public:
  virtual void  update_motor(sensory_sys*);
  virtual char* name() { return "Wall_seek"; }
  virtual char* label() { return "Ws"; }   
  virtual char* mode_name();
  wall_seek(int id) : behaviour(id) {}
};


void wall_seek::update_motor(sensory_sys *senses) 
{
  int sumd, diff, side;
  float lws, rws, ws;
  
  aBusy = 0.0;

  sumd= senses->infSum();
  diff= senses->infDiff();
  side= senses->infNearSide();

  if (sumd > 500)                                      // obstacle avoidance 
    {
      ws= sws(diff);
      switch (side) {
      case LEFT:  lws= ws; rws= -ws; aMode= 0; break;  // turn right
      case RIGHT: lws= -ws; rws= ws; aMode= 1;         // turn left
      }
    }
  
  else if (sumd > 10)                                  // slow ahead   
    { lws= rws= 0.55; aMode= 2; }
  
  else                                                 // full ahead 
    { lws= rws= 1.0; aMode= 3; }
  
  aMotor[eLws]= clip_motor(lws);
  aMotor[eRws]= clip_motor(rws);
  aMotor[eA]= (POSTURE)? -1.0 : 0.0;
  aMotor[eG]= 0.0;
  get_vector();
}

char* wall_seek::mode_name()
{
  static char *modes[] = {"turn right", 
			  "turn left",  
			  "slow ahead", 
			  "full ahead"};

  return modes[aMode];
}



// nest seek class termed WALL-FOLLOW in publications

class nest_seek : public behaviour
{
  int aLastSide;

 public:
  virtual void  update_motor(sensory_sys*);
  virtual char* name() { return "Nest_seek"; }
  virtual char* label() { return "Ns"; }   
  virtual char* mode_name();
  nest_seek(int id) : behaviour(id) { aLastSide= LEFT; }
};


void nest_seek::update_motor(sensory_sys *senses) 
{
  int sumd, side; 
  float lws, rws, slower, faster, vslow= 0.15;         // 0.15
  int prefd= 1200; 
  // double tim= aTimer.check();

  aBusy = 0.0;

  sumd= senses->infSum();
  side= senses->infNearSide();
  slower= 0.4 - 0.00040*fabs(sumd - prefd);             // was 0.4 - 0.00050
  faster= 0.4 - 0.00035*fabs(sumd - prefd);             // was 0.4 - 0.00035
  
  if (sumd > 2000)                                      // much too close to wall
    {
      switch (side) {
      case LEFT:  lws= vslow; rws= -vslow; break;       // rotate right
      case RIGHT: lws= -vslow; rws= vslow;              // rotate left
      }
      aMode= 0;
    }

  else if (sumd >= prefd)                               // close to wall or optimal
    {
      switch (side) {
      case LEFT:  lws= faster; rws= slower; break;      // veer right or straight-ahead
      case RIGHT: lws= slower; rws= faster;             // veer left or straight-ahead
      }
      aMode= 1;
    }

  else if (sumd > 600)                                  // too far away
    {
      switch (side) {
      case LEFT:  lws= slower; rws= faster; break;      // veer left
      case RIGHT: lws= faster; rws= slower;             // veer right
      }
      aMode= 2;
    }
  
  else                                                  // much too far away
    {
      switch (side) {
      case LEFT:  lws= 0.20; rws= 0.27; break;          // veer left
      case RIGHT: lws= 0.27; rws= 0.20;                 // veer right
      }
      aMode= 3;
    }

  if (sumd>20) aLastSide= side;

  // generate a busy signal if losing contact with the wall 
  if (senses->infObs()==1) aBusy= 1.0;

  aMotor[eLws]= lws;
  aMotor[eRws]= rws;
  aMotor[eA]= (POSTURE)? -1.0 : 0.0;
  aMotor[eG]= 0.0;

  get_vector();
}

char* nest_seek::mode_name()
{
  static char *modes[4] = {"Rotate",
			   "Out/Adv",
			   "In", 
			   "In sharp"};

  return modes[aMode];
}



// can deposit class 

class can_deposit : public behaviour
{
 public:
  virtual void  update_motor(sensory_sys*);
  virtual char* name() { return "Can_deposit"; }
  virtual char* label() { return "Cd"; }   
  virtual char* mode_name();    
  can_deposit(int id) : behaviour(id) {}
};

void can_deposit::update_motor(sensory_sys *senses)
{
  // these are the times in seconds for each stage of the action pattern 
  const double app_light= 0.8, arm_down= 0.8, open_grip= 0.8, arm_up= 0.8;
  float lws, rws, arm, grip;
  double tim= aTimer.check();

  aBusy= 1.0;
  if (tim < app_light)                            // go toward the centre of the nest
    {
      aMode= 0; arm= 0.0; grip= 0.0; aBusy= 0.0;
      switch (senses->ambLightSide()) {
      case LEFT:  lws= 0.14; rws= 0.27; break;    // veer left
      case RIGHT: lws= 0.27; rws= 0.14;           // veer right
      }
    }
  else if (tim < app_light+arm_down)                   // lower the arm 
    { aMode= 1; lws= rws= 0; arm= 0.5; grip= 0.0; }
  
  else if (tim < app_light+arm_down+open_grip)    // release the can 
    { aMode= 2; lws= rws= 0;
      arm= (POSTURE)? 0.5 : 0.0; 
      grip= -1.0; }
  
  else if (tim < app_light+arm_down+open_grip+arm_up) // raise the arm 
    { aMode= 3; lws= rws= 0.0; arm= -1.0; grip= 0.0; }

  else                                             // finished- set the busy signal to zero 
    { 
      aMode= 4; lws= rws= grip= 0.0; 
      arm= (POSTURE)? -1.0 : 0.0;
      aBusy= 0.0;
      reset();
    }

  aMotor[eLws]= lws;
  aMotor[eRws]= rws;
  aMotor[eG]= grip;
  aMotor[eA]= arm;
  get_vector();
}

char* can_deposit::mode_name()
{
  static char *modes[] = {"approach",
			  "arm down", 
			  "gripper open", 
			  "arm up", 
			  "finished"}; 

  return modes[aMode];
}


#endif







