/*****************************************************************************
 * File:         motivation.h                                                *
 * Version:      4.0                                                         *
 * Description:  motivational sub-systems                                    *
 * Authors:      Tony Prescott and Fernando Montes                           *
 * Contact:      t.j.prescott@shef.ac.uk                                     *
 *****************************************************************************/

// *** major update to introduce 'mood' control
// a 'mood' is a pair of multiplicative and additive factors that are used
// to adjust the value of the motivation at each time-step
// the SET_MOOD method is used to switch between moods

#ifndef MOTIVATION
#define MOTIVATION


#include <stream.h>
#include <vector>
#include <string>

class motivation;
typedef vector <motivation*> mvector;
typedef pair <float, float> mmode;
typedef vector <mmode> mmvector;

class motivation : public subsys {
 protected:
  bool  aDynamic;                         // true if time-varying, false otherwise
  float aInit_value;                      // initial value
  float aMin;                             // minimum value
  float aMax;                             // maximum value
  int aMood;                              // current 'mood' in which the motivation is operating
  mmvector aMoods;                        // vector of multiplicative and additive factors 
                                          // corresponding to each of the  moods of the motivation
  stat aMstat;                            // maintains statistics of changing value over time
 public:
  void set_range(float min, float max) { aMin= min, aMax= max; }
  int set_dynamic(bool flag) { return aDynamic= flag; }
  int add_mood(float mult, float add)
    {
      aMoods.push_back(make_pair(mult, add)); return aMoods.size();
    }
  int set_mood(int m) { return aMood= m; }
  float init_value(float v) { aInit_value= v; set_value(v); aVtm1= v; aMstat.zero(); return v; }
  float reset() { return init_value(aInit_value); }
  float mean() { return aMstat.mean(); }
  motivation(char *name) : subsys (name, NULL), aDynamic(false), aMax(1.0) {}
  
  float set_value(float);
  virtual float update();
  void print(int);
};


float motivation::set_value(float v) 
{
  aVtm1= aV;
  aV= v;
  // printf("%s: set value t %f t-1 %f\n", name(), aV, aVtm1);
  if (aV > aMax) aV= aMax;
  else if (aV < aMin) aV= aMin;
  return aV;
}


float motivation::update()
{
  // cout << aName << "::update\n";
  if (aDynamic)
    {
      // cout << "direction " << aDirection << "\n";
      set_value (aV*aMoods[aMood].first + aMoods[aMood].second);
    }
  aMstat.add_value(aV);
  return (aV);
} 

void motivation::print(int verbose)
{
  cout << name();
  if (verbose>0) {
    if (aDynamic) {
      cout << "(" << aMin << ":" << aMax << " Moods : ";
      for (uint i=0; i<aMoods.size(); i++) 
	cout << aMoods[i].first << "|" << aMoods[i].second << " ";
    }
    else
      cout << "(" << aV << ")";      
  } 
} 

/* Fear:                                                                    *
 * check if fear reduction conditions are satisfied (away from open spaces) *
 * Then reduce or increase value  multiplicatively as appropriate.          */

class fear_ms : public motivation {
  percept *aWallPtr;
  percept *aNestPtr;
 public:
  fear_ms(percept *wp, percept *np) : motivation("Fear"), aWallPtr(wp), aNestPtr(np) {}
  virtual float update();
};
  
float fear_ms::update()
{
  if (aDynamic)
    {
      aMood= (aWallPtr->value()==1 || aNestPtr->value()==1)? 0 : 1;
      motivation::update();
    }
  return (aV);
}


/* Hunger:                                                                  *
 * check if hunger reduction conditions satisfied (can deposited in nest)   * 
 * Reduced or increased additively when as appropriate                      */

class hunger_ms : public motivation {
  percept* aGripPtr;
  percept* aNestPtr;
 public:
  hunger_ms(percept *gp, percept *np) : motivation("Hunger"), aGripPtr(gp), aNestPtr(np) {}
  virtual float update();
};

float hunger_ms::update()
{
  if (aDynamic)
    {
      aMood= (aGripPtr->valuetm1()==1 && 
	      aGripPtr->value()==-1 && 
	      // aNestPtr->value()==1
	      aNestPtr->recency()<=10
	      )? 
	1 : 0;
      motivation::update();
    }
  return aV;
}

#endif



