/*****************************************************************************
 * File:         percept.h                                                   *
 * Version:      4.0                                                         *
 * Description:  perceptual sub-systems                                      *
 * Authors:      Tony Prescott and Fernando Montes                           *
 * Contact:      t.j.prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef PERCEPT
#define PERCEPT

#include <gui.h>

#include <stream.h>
#include <string.h>

/* defines the class PERCEPT of perceptual sub-systems and                   *
 * specialised sub-classes WALL_PS, GRIP_PS, CAN_PS, CORNER_PS,              *
 * LIGHT_PS, and NEST_PS.                                                    *
 * Note that NEST_PS differs from the others in that it has inputs           *
 * (attributes) that are other perceptual sub-systems                        *
 * Percept values are currently -1 or +1 however floats are used to          *
 * to allow future extension to real valued percepts                         */


class percept : public subsys {
  int aRecency;
 public:
  virtual float update()=0;
  int inc_recency(bool flag) 
    { 
      if (flag) aRecency= 0; 
      else if (aRecency<INT_MAX) aRecency++;
      return aRecency;
    }
  int recency() { return aRecency; }
  percept(char *name, sensory_sys *sp, float init_v =-1.0) 
    : subsys(name, sp, init_v), aRecency(INT_MAX) {}
};

typedef vector <percept*> pvector;


/* wall_ps                                                               *
 * EITHER the sum of forward input must exceed WALL_THRESH AND           *
 * at least three detectors must sense obstacles (ie. exceed OBS_THRESH  *
 * OR either the leftmost (0) or rightmost (5) sensor must itself exceed *
 * WALL_THRESH. The arm cannot be down.                                  */

class wall_ps : public percept {
 public:
  virtual float update();
  wall_ps(sensory_sys* r) : percept("Wall", r) {}
};

float wall_ps::update()
{
  if (((aSptr->infSum()>WALL_THRESH && aSptr->infObs()>=3) ||
       aSptr->inf(0)>WALL_THRESH || aSptr->inf(5)>WALL_THRESH))
    {
      if (aSptr->armDown()==1) set_value(-1.0);
      //if (aSptr->armUp()!=1) set_value(-1.0); // arm in the way- ambiguous reading
      else set_value(1.0);
    }
  else set_value(-1.0);
  // if (change()) gui_beep();
  return value();
}


/* grip_ps                                                               *
 * Looks for an object in the gripper. Currently uses both the           *
 * object presence and resistivity sensors (i.e true if either is 1.0)   */

class grip_ps : public percept {
 public:
  virtual float update();
  grip_ps(sensory_sys* r) : percept("Grip", r) {}
};

float grip_ps::update()
{  
  if ((aSptr->gripDetect()==1.0 || 
       (USE_RESISTANCE && aSptr->gripResist()==1.0)))
    //  && aSptr->gripPos()>5)               // not shut and empty
    return set_value(1.0);
  else return set_value(-1.0);
}

 
/* can_ps                                                                *
 * Looks for a nearby object on the two foremost sensors (2 and 3)       *
 * and empty space either side (1 and 4). Disabled in the presence       *
 * of strong light (which confuses the ir sensors                        */

class can_ps : public percept {
 public:
  virtual float update();
  can_ps(sensory_sys* r) : percept("Can", r) {}
};

float can_ps::update()
{  
  if (aSptr->inf(2)>CAN_FRONT && aSptr->inf(3)>CAN_FRONT &&
      aSptr->inf(1)<CAN_SIDE && aSptr->inf(4)<CAN_SIDE &&
      aSptr->ambLit()==0)
    {
      // cout << "can detected\n";
      return set_value(1.0);
    }
  else return set_value(-1.0);
}


/* corner_ps                                                              *
 * The number of obstacles detected must equal or exceed CORNER_OBS       *
 * AND the total IR activity must exceed CORNER_THRESH. Disabled when     *
 * the arm is down (which is otherwise detected as a corner). This is     *
 * not used in the current implementation as its rather unreliable        */

class corner_ps : public percept {
 public:
  virtual float update();
  corner_ps(sensory_sys* r) : percept("Corner", r) {}
};

float corner_ps::update()
{
  if (aSptr->infObs()>=CORNER_OBS && 
      aSptr->infSum()>CORNER_THRESH &&
      aSptr->armDown()==0)
    return set_value(1.0);
  else return set_value(-1.0);
}


/* light_ps                                                               *
 * Detects the light source using the ambient light values of the sensor  *
 * array.  The nest is detected when activity values of at least two of   *
 * the six forward sensors are lit (i.e. below LIGHT_THRESH)              */

class light_ps : public percept {
 public:
  virtual float update();
  light_ps(sensory_sys* r) : percept("Light", r) {}
};
 

float light_ps::update()
{
  if (aSptr->ambLit()>=LIGHT_COUNT) return set_value(1.0);
  else return set_value(-1.0);
}


/* nest_ps                                                                *
 * Detecting a nest depends on a combination of                           *
 * detecting a light, a corner, and a wall                                */

class nest_ps : public percept {
  percept *aLightPtr;
  percept *aCornerPtr;
  percept *aWallPtr;
 public:
  virtual float update();
  nest_ps(percept* lp, percept* cp, percept* wp) : percept("Nest", NULL),
    aLightPtr(lp), aCornerPtr(cp), aWallPtr(wp) {}
};

float nest_ps::update()
{
  if ((USE_LIGHT==0 || aLightPtr->value()==1.0) &&
      (USE_CORNER==0 || aCornerPtr->value()==1.0) &&
      (USE_WALL==0 || aWallPtr->value()==1.0))
    {
      inc_recency(true);
      return set_value(1.0);
    }
  else 
    {
      inc_recency(false);
      return set_value(-1.0);
    }
}



#endif














