/*****************************************************************************
 * File:         sensory_sys.h                                               *
 * Version:      4.0                                                         *
 * Description:  Processes all the raw sensory data from the robot           *
 * Authors:      Tony Prescott and Fernando Montes                           *
 * Contact:      t.j.prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef SENSORY_SYS
#define SENSORY_SYS


#include <stream.h>
#include <stdio.h>
#include <Khepera.h>
#include <KheperaGripper.h>

class sensory_sys {
  enum Motor_vals {eUp= 152, eMiddle= 227, eDown= 255, eOpen= 0, eClosed= 1 }; 
  int aInf[N_IR];            // Infra-red sensor data, 0(~30mm)-1023+(0mm) 
  int aInfSum;               // total forward sensor values 0-5
  int aInfLeft;              // total left sensor values 0-2
  int aInfRight;             // total right sensor value 3-5
  int aInfObs;               // number of forward sensors > OBS_THRESH
  int aInfNearSide;          // which of left and right has highest total
  int aInfDiff;              // absolute difference between left and right

  int aAmb[N_IR];            // Ambient light sensor data, 50(light)-450(dark)
  int aAmbSum;               // total forward light values
  int aAmbLeft;              // total left light values 0-2
  int aAmbRight;             // total right light values 3-5
  int aAmbLit;               // number of forward sensors < LIGHT_THRESH
  int aAmbLightSide;         // which or left and right is more brightly lit
  int aAmbDiff;              // absolute difference between left and right

  int aGripPos;              // Gripper data
  int aGripDetect;           //
  int aGripResist;     
  
  int aArmPos;               // Arm data
  
 public:
  int inf(int n) { return aInf[n]; }
  int infSum() {return aInfSum; }
  int infLeft() {return aInfLeft; }
  int infRight() {return aInfRight; }
  int infDiff() { return aInfDiff; }
  int infObs() { return aInfObs; }
  int infNearSide() {return aInfNearSide; }
  
  int amb(int n) { return aAmb[n]; }
  int ambSum(){ return aAmbSum; }
  int ambLeft() { return aAmbLeft; }
  int ambRight() { return aAmbRight; }
  int ambLit() { return aAmbLit; }
  int ambLightSide() { return aAmbLightSide; }
  int ambDiff() { return aAmbDiff; }
  
  int gripPos() { return aGripPos; }
  int gripDetect() { return aGripDetect; }
  int gripResist() { return aGripResist; }
  int armDown() { return (aArmPos>eMiddle)? 1 : 0; }
  int armUp() { return (aArmPos<(eUp+10))? 1 : 0; }
  void print_senses();
  void update();
};
 
void sensory_sys::print_senses()
{
  printf("Infra-red: sum %3d, left %3d, right %3d, obs %1d, Nearside %1d, diff %3d\n",
	 aInfSum, aInfLeft, aInfRight, aInfObs, aInfNearSide, aInfDiff);
  printf("Ambient:   sum %3d, left %3d, right %3d, lit %1d, Lightside %1d, diff %3d\n",
	 aAmbSum, aAmbLeft, aAmbRight, aAmbLit, aAmbLightSide, aAmbDiff);
  printf("Gripper:   Position %d Detect %d, Resist %d\n", aGripPos, 
	 aGripDetect, aGripResist);
  printf("Arm:       Position %d, Down %d\n\n", aArmPos, armDown());
}

void sensory_sys::update()
{
  int i, d, l;

  // get Infra-red sensor values

  aInfSum= aInfLeft= aInfRight= aInfObs= 0;
  for (i=0; i<N_IR; i++)
    {
      aInf[i]= d= khepera_get_proximity(i);
      if (i < 6) 
	{
	  aInfSum += d;
	  if (i < 3) aInfLeft += d;
	  else aInfRight += d;
	  if (d > OBS_THRESH) aInfObs++;
	}
      aInfNearSide= (aInfLeft >= aInfRight)? LEFT : RIGHT;
      aInfDiff= abs(aInfLeft - aInfRight);
    }

  // get Ambient light sensor values
  
  aAmbSum= aAmbRight= aAmbLeft= aAmbLit= 0;
  for (i=0; i<N_IR; i++)
    {
      aAmb[i]= l= khepera_get_light(i);
      if (i < 6)
	{
	  aAmbSum+= d;
	  if (i < 3) aAmbLeft += l;
	  else aAmbRight += l;

 	  if (l < LIGHT_THRESH) aAmbLit++;
	}
      aAmbLightSide= (aAmbLeft <= aAmbRight)? LEFT : RIGHT;
      aAmbDiff= abs(aAmbLeft - aAmbRight);
    }

  // get Gripper and Arm values
  aGripDetect= (gripper_get_presence() == true)? 1 : 0;
  aGripResist= (gripper_get_resistivity() > CAN_RESISTIVITY)? 1 : 0;
  aGripPos= gripper_get_grip();
  aArmPos= gripper_get_arm();      

  // if (aGripPos<100) cout << "gripper " << aGripPos << endl;
  // print_senses();
}

#endif












