/*****************************************************************************
 * File:         graph_window.h                                              *
 * Version:      4.0                                                         *
 * Description:  window class for displaying graphs                          *
 * Authors:      Tony Prescott and Fernando Montes                           *
 * Modified:     T.J.Prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef GRAPH_WINDOW
#define GRAPH_WINDOW

#include "iwindow.h"

#define GUI_GREY gui_color_get(200,204,200)
#define MAX_T 300

class graph_window : public iwindow
{
  bool          aSmall_size;
  bool          aUse_time;
  int           aWx;
  int           aWy;
  int           aWw;
  int           aWh;
  int           aFx;
  int           aFy;
  int           aFw;
  int           aFh;
  double        aMax_time;
  int           aXtm1;
  
  int get_x(double t, int off= 0) { return off + (int) floor((t/aMax_time) * aFw); }
  int get_y(float v) { return (int) (aWh - (aFy + 3) - v * (aFh - 3)); }
  int draw_graph(float, float, uint32);
  int draw_graph(double, float, float, uint32);
 public:
  void create(bool, bool, double);
  void remake(bool);
  void remake();
  void update(mvector&, double);
  virtual char* name() { return "graph"; }
  graph_window() : iwindow() {}
};

void graph_window::create(bool small_size= true, bool use_time= true, double max_t= MAX_T)
{
  if (!aExists)
    {
      aSmall_size= small_size;

      if (aSmall_size)
	{
	  aWx= 0; aWy= 10;      // x, y position
	  aWw= 410; aWh= 190;   // width, height
	  aFx= 30;  aFy= 20;    // border width, height
	}
      else
	{
	  aWx= 105; aWy= 10;     // x, y position
	  aWw= 818; aWh= 90;    // width, height
	  aFx= 20;  aFy= 10;    // border width, height
	}
      
      aUse_time= use_time;
      aMax_time= max_t;

      aFw= aWw - (2*aFx);
      aFh= aWh - (2*aFy);

      aWindow= gui_window_new(name(), aWx, aWy, aWw, aWh);
      
      gui_line_new(aWindow, aFx, aFy+aFh, aFx+aFw, aFy+aFh, GUI_BLACK);
      gui_line_new(aWindow, aFx, aFy, aFx, aFy+aFh, GUI_BLACK);
      aWindow_frame= gui_rectangle_new(aWindow, aFx+1, aFy, aFw, aFh, GUI_GREY, true);

      for (int i= 0; i<10; ++i)
	{
	  int y= get_y((1+i) * 0.1);
	  int w= ((i+1)%5==0)? 10 : 5; 
	  gui_line_new(aWindow, aFx, y, aFx - w, y, GUI_BLACK);
	  gui_line_new(aWindow, aFx, y + 1, aFx - w, y + 1, GUI_BLACK);
	}

      gui_window_show(aWindow);
      aExists= 1;
      aWindow_clock= 0;
      aXtm1= 0;
    }
}

void graph_window::remake(bool small_size)
{
  if (aExists)
    {
      // cout << "remaking graph window\n";
      gui_window_delete(aWindow);
      aWindow= NULL;
      aExists= 0;

      create(small_size);
    }
}

void graph_window::remake() { remake(aSmall_size); }

/* show graph window */

int graph_window::draw_graph(float prevalue, float value, uint32 color)
{
  int x, y1, y2;
  gui_window w= aWindow;

  x= aWindow_clock + aFx + 1;
  y1= get_y(prevalue);
  y2= get_y(value);
  gui_line_new(w, x, y1, x, (int) y2, color);  
  gui_line_new(w, x, y1+1, x, y2+1, color);
  gui_line_new(w, x, y1+2, x, y2+2, color);
  
  return 1;
}

int graph_window::draw_graph(double t, float prevalue, float value, uint32 color)
{
  const int y_offset= aWh - (aFy + 3), y_scale= aFh - 3; //***
  
  if (aExists)
    { 
      int xt= get_x(t), x1, x2, y1, y2;

      if (xt >= aFw)
	{
	  xt= xt % aFw;
	  if (xt < aXtm1) remake();  // was clear();
	  // cout << "Clear graph\n";
	  aXtm1= xt;
	}
	  
      y1= (int) (y_offset - prevalue * y_scale);
      y2= (int) (y_offset - value * y_scale);
      x1= aWindow_clock + aFx + 1;
      x2= xt + aFx + 1;
      gui_line_new(aWindow, x1, y1, x2, y2, color);  
      gui_line_new(aWindow, x1, y1+1, x2, y2+1, color);
      gui_line_new(aWindow, x1, y1+2, x2, y2+2, color);
      return xt - aWindow_clock;
    }
  return 0;
}

// update graph window 

void graph_window::update(mvector& motives, double et= -1.0)
{
  const uint32 s_colors[]= {GUI_BLUE, GUI_MAGENTA, GUI_WHITE, GUI_GREEN, GUI_WHITE, 
			    GUI_RED, GUI_CYAN, GUI_BLACK, GUI_RED };

  if (aExists)
    {
      if (aUse_time && et>=0.0)
	{
	  // cout << "max t " << aMax_time << endl;
	  int d;
	  for (uint i=0; i<motives.size(); i++)
	    d= draw_graph(et, motives[i]->valuetm1(), motives[i]->value(), s_colors[i]);
	  aWindow_clock+= d;
	}
      else
	{
	  if (aWindow_clock % aFw == 0) remake();             // reached end of frame 
	  
	  for (uint i=0; i<motives.size(); i++)
	    draw_graph(motives[i]->valuetm1(), motives[i]->value(), s_colors[i]);
	  aWindow_clock++;
	}
    }
}

#endif

















