/*****************************************************************************
 * File:         interface.h                                                 *
 * Version:      3.0                                                         *
 * Description:  class for managing all windows                              *
 * Authors:      Tony Prescott                                               *
 * Modified:     T.J.Prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef INTERFACE
#define INTERFACE  

// INTERFACE CLASS

#include "iwindow.h"           // base gui window class
#include "tools_window.h"      // window for online controls                                
#include "percepts_window.h"   // window to display perceptual variables
#include "graph_window.h"      // window to display motivations
#include "ethogram_window.h"   // window to display behavioural selections
#include "monitor_window.h"    // window to show activity of basal ganglia nuclei
#include "salience_window.h"   // window to display channel salience
#include "time_bar.h"          // shows elapsed time as a thermometer bar

#include "observer_log.h"      // class for generating Observer Pro log files  
#include <stdio.h>

class interface
{
 private:
  /* windows */
  bool            aFull_display;
  bool            aTools_flag;
  tools_window    aTools_w;

  bool            aPercepts_flag;
  percepts_window aPercepts_w;

  bool            aGraph_flag;
  graph_window    aGraph_w;

  bool            aEthogram_flag;
  ethogram_window aEthogram_w;

  bool            aSalience_flag;
  salience_window aSalience_w;

  bool            aMonitor_flag;
  monitor_window  aMonitor_w;

  bool            aTime_bar_flag;
  time_bar        aTime_bar;

 public:
  void beep() { gui_beep(); }

  void callback(flag_vector& flags) { 
    if (gui_event_get_info() == GUI_DOUBLE_CLICK)
      {
	// cout << "gui  double click\n";
	// cout << "key event: " << gui_event_get_key() << endl;
	beep();
	aGraph_w.clear();
	aEthogram_w.clear();
	aTime_bar.clear();
    }
    else aTools_w.callback(flags); 
  }
  void set_button_enabled(int i, bool flag) { aTools_w.set_button_enabled(i, flag); }
  void set_button_label(int i, char s[]) { aTools_w.set_button_label(i, s); }
  // ***
  void update_button(gui_flag &f) { aTools_w.set_button_label(f.id(), f.label()); beep(); }

  // these functions all simply pass on their parameters to the relevant window
  void update_percepts_window(pvector& ps) { aPercepts_w.update(ps); }
  void update_graph_window (mvector& ms, double et= -1.0) { aGraph_w.update(ms, et); }
  void init_graph_window() { aGraph_w.remake(); }
  void update_ethogram_window(bvector& bs) { aEthogram_w.update(bs); }
  void init_ethogram_window(bvector& bs) { aEthogram_w.remake(bs); }
  void update_salience_window(bvector& bs, float *v1, float *v2) { aSalience_w.update(bs, v1, v2); }
  // void update_salience_window(float *v, int c) { aSalience_w.update(v, c); }
  void update_monitor_window() { aMonitor_w.update(); }
  void update_time_bar(float e, int w= -1) { aTime_bar.update(e, w); }
  void init_time_bar(float e) { aTime_bar.set_max_time(e); }
  void countdown_time_bar(float e, float me) { aTime_bar.countdown(e, me); }
  void make_windows(string, flag_vector&, pvector&, bvector&, basal_ganglia*);
  void remake_windows(string, flag_vector&, pvector&, bvector&, basal_ganglia*);
  string& efolder_name() { return aTools_w.efolder_name(); }
  interface(bool full= true)
    {
      aFull_display= full;
      aTools_flag= true;
      aPercepts_flag= true;
      aGraph_flag= true;
      aEthogram_flag= true;
      aSalience_flag= true; // problem window! 1.4.03
      aMonitor_flag= true;
      aTime_bar_flag= true;
    }
};


// Main function for creating, deleting, or recreating windows

void interface::make_windows(string cname, flag_vector& flags, 
			     pvector& percepts, bvector& behaviours, 
			     basal_ganglia *bg)
{
  const int verbose= 0;

  if (aTools_flag) {
    if (verbose) cout << "creating TOOLS\n";
    aTools_w.create(cname, flags);
  }
  if (aPercepts_flag && aFull_display) {
    if (verbose) cout << "creating PERCEPTS\n";
    aPercepts_w.create(percepts);
  }
  if (aGraph_flag) {
    if (verbose) cout << "creating GRAPH\n";
    aGraph_w.create(aFull_display);
  }
  if (aEthogram_flag && aFull_display) {
    if (verbose) cout << "creating ETHOGRAM\n";
    aEthogram_w.create(behaviours);
  }
  if (aSalience_flag && aFull_display) {
    if (verbose) cout << "creating SALIENCE\n";
    aSalience_w.create(behaviours);
  }
  if (aMonitor_flag && aFull_display)  {
    if (verbose) cout << "creating MONITOR\n";
    aMonitor_w.create(bg);
  }
  if (aTime_bar_flag) {
    if (verbose) cout << "creating TIME BAR\n";
    aTime_bar.create(aFull_display);
  }
}

void interface::remake_windows(string cname, flag_vector& flags, 
			       pvector& percepts, bvector& behaviours, 
			       basal_ganglia *bg)
{
  const int verbose= 0;

  /*
  if (aTools_flag) {
     if (verbose) cout << "remaking TOOLS\n";
    aTools_w.remake(cname, flags);
  }
  */

  if (aPercepts_flag && aFull_display) {
    if (verbose) cout << "remaking PERCEPTS\n";
    aPercepts_w.remake(percepts);
  }

  if (aGraph_flag) {
    if (verbose) cout << "remaking GRAPH\n";
    aGraph_w.remake(aFull_display);
  }
  if (aEthogram_flag && aFull_display) {
    if (verbose) cout << "remaking ETHOGRAM\n";
    aEthogram_w.remake(behaviours);
  }

  if (aSalience_flag && aFull_display) {
    if (verbose) cout << "remaking SALIENCE\n";
    aSalience_w.remake(behaviours);
  }

  if (aMonitor_flag && aFull_display)  {
    if (verbose) cout << "remaking MONITOR\n";
    aMonitor_w.remake(bg);
  }

  if (aTime_bar_flag) {
    if (verbose) cout << "remaking TIME BAR\n";
    aTime_bar.remake(aFull_display);
  }
}


// controller_gui method using interface

int last_it= -1;

bool controller_gui::show_etime(gui_flag& gf, float elapsed, int maxt, int ms= 0, 
				char* pref= "", char mark= '.')
{
  int et= (int) floor(elapsed);
  int it= (maxt!=0)? maxt - et : et;
  
  if (it!= last_it)
    {
      if (ms != 0 && et % ms == 0 && et != 0) 
	{
	  if (et % 15 == 0) cout << et; 
	  else cout.put(mark);
	  if (et % (30 * ms) == 0) cout << endl;
	  else cout.flush();
	}
      char s[100];
      if (maxt==0) sprintf(s, "%s %i", pref, it);
      else sprintf(s, "%s %i (%i)", pref, it, maxt);
      // cout << s << gf.id() << endl;
      aInterface->set_button_label(gf.id(), s);
      last_it= it;
      return true;
    }
  else return false;
}

#endif











