/*****************************************************************************
 * File:         monitor_window.h                                            *
 * Version:      4.0                                                         *
 * Description:  window class for showing bg nuclei activation               *
 * Authors:      Tony Prescott and Fernando Montes                           *
 * Modified:     T.J.Prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef MONITOR_WINDOW
#define MONITOR_WINDOW

#include "iwindow.h"

#define MAX_CHAN 10
#define MROWS     2
#define MCOLS     5

using namespace bgs;

class monitor_window : public iwindow
{
 private:
  const int aRows;
  const int aCols;
  const int aMax_channels;
  gui_rectangle aHist_rects[MROWS][MCOLS][MAX_CHAN];

  basal_ganglia* aBg;
  int aN_channels;
  int aWindow_clock;
  int aN_cells;
  int aCell_height;
  int aCell_width;
  int aFx;
  int aFy;
  int aHist_height;
  int aHist_width;
  int aHist_y_offset;
 public:
  void remake(basal_ganglia*);
  void create(basal_ganglia*);
  gui_rectangle draw_hist(int, int, int, float);
  void update();
  monitor_window() : aRows(MROWS), aCols(MCOLS), aMax_channels(MAX_CHAN) {}
};


gui_rectangle monitor_window::draw_hist(int row, int col, int channel, float value)
{
  const int offset=30;
  const float max_val = 1.0;
  int x, y, height;
  uint32 color;
  
  if (value>max_val) value= max_val;
  else if (value<-max_val) value= -max_val;

  height= (int) (value * aHist_height);
  
  if (height > 0) color= GUI_RED;
  else if (height < 0) { color= GUI_BLUE; height= -height; }
  else { color= GUI_BLACK; height= 1; }

  x= offset + (col * aCell_width) + (channel + 1) * aHist_width;
  y= offset + ((row + 1) * aCell_height) - (aHist_y_offset + height);

  return gui_rectangle_new(aWindow, x, y, aHist_width, height, color, true);
}

void monitor_window::remake(basal_ganglia *bg= NULL)
{
  if (aExists)
    {
      cout << "remaking monitor window\n";
      gui_window_delete(aWindow);
      aWindow= NULL;
      aExists= 0;

      create(bg);
    }
}

void monitor_window::create(basal_ganglia *bg= NULL)
{
  const int wx= 420, wy= 10, ww= 400, wh= 260;
  int i, j, k, x, y, y_offset, f_width, f_height;

  if (!aExists)
    {
      if (bg!=NULL) aBg= bg;
      // cout << "Creating monitor window for " << aBg->name() << " with "; 
      // cout << aBg->n_nuclei() << " nuclei and " << aBg->n_channels() << " channels\n";
      aN_channels= aBg->n_channels();

      if (aN_channels>aMax_channels) 
	cout << "warning: two many channels for window " << name();
      aN_cells= aCols * aRows;
      
      aFx= 30;
      aFy= 20;
      y_offset= wh - aFy;
      f_width= ww - (2 * aFx);
      f_height= wh - (2 * aFy);
      aCell_height= (int) f_height / aRows;
      aCell_width= (int) f_width / aCols;
      aHist_width= (int) aCell_width / (aN_channels + 2);
      aHist_height= (int) (aCell_height * 2) / 3;
      aHist_y_offset= (int) (aCell_height - aHist_height) / 2;
      
      aWindow= gui_window_new("monitor", wx, wy, ww, wh);
      
      x= aFx; y= aFy;
      for (i=0; i<(aRows + 1); i++)
	{
	  gui_line_new(aWindow, aFx, y, aFx + f_width, y, GUI_BLACK);
	  y += aCell_height;
	}
      
      for (i=0; i<(aCols + 1); i++)
	{
	  gui_line_new(aWindow, x, aFy, x, aFy + f_height, GUI_BLACK);
	  x += aCell_width;
	}
      
      for (i=0; i<aRows; i++)
	{
	  for (j=0; j<aCols; j++)
	    {
	      gui_label_new(aWindow, aFx + (j * aCell_width) + aHist_width,
			    aFy + (i * aCell_height) + aHist_y_offset - 5, 
			    aBg->nucleus_name(i*aCols + j));
	    
	      for (k=0; k<aN_channels; k++)
		aHist_rects[i][j][k]  = draw_hist(i, j, k, 0.0);
	    }
	}
      gui_window_show(aWindow);
      aExists= 1;
      aWindow_clock= 0;
    }
}

void monitor_window::update()
{ 
  // const int refresh= 250;
  int row, col, i;
  Leaky_value p= eY;           // enum  Leaky_value {eU= 0, eA= 1, eY= 2}
  nvector& nuclei= aBg->nuclei();
  nucleus *nuc;

  if (aExists)
    { 
      // if (aWindow_clock==refresh) remake();

      for (uint cell=0; cell<nuclei.size(); cell++)
	{
	  if (cell< (uint) aN_cells) 
	    {
	      row= cell / aCols;
	      col= cell % aCols;
	      nuc= nuclei[cell];

	      for (i=0; i<aN_channels; i++)
		{
		  gui_rectangle_delete(aHist_rects[row][col][i]);
		  aHist_rects[row][col][i]= draw_hist(row, col, i, nuc->value(p, i));
		}
	    }
	}
      aWindow_clock++;
    }
}

#endif








