/*****************************************************************************
 * File:         utils/bv_stat.h                                             *
 * Version:      1.0                                                         *
 * Description:  Stores contour plot data for a bivariate variable           *
 * Author:       Tony Prescott                                               *
 * Contact:      T.J.Prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef UTILS_BV_STAT
#define UTILS_BV_STAT

#include <cstdlib>
// #include <stream.h>
#include <math.h>
#include <vector>
#include <algorithm>
#include <string>
#include "io.h"

typedef vector <double> db_vector;

class bv_stat {
 private:
  char*  aName;
  int    aCount;
  double aV1;
  double aV2;
  int    aDim;
  double aMinv;
  double aMaxv;
  double aMult;
  int* aSurf;
  float axis_label(int c);
  bool check_sum();
 public:
  void init(char*);
  void zero();
  bv_stat(char*, int, double, double);
  double add_value(double, double);
  void set_name(char*, char* suffix= NULL);
  void print(bool, bool);
  void print_cols(bool, bool);
};

bv_stat::bv_stat(char* name= NULL, int dim= 10, double minv= 0.0, double maxv= 1.001)
{
  set_name(name);
  // cout << "new stat '" << name << "' save ";
  // if (save) cout << "enabled\n";
  // else cout << "disabled\n";
  aDim= dim;
  aMinv= minv;
  aMaxv= maxv;
  aMult= dim/(maxv - minv);
  aSurf= new int[dim*dim];
  zero();
}

void bv_stat::zero() 
{
  // cout << "initialising " << aName << "\n";
  aV1= 0.0;
  aV2= 0.0;
  aCount= 0;
  for (int i=0; i<aDim*aDim; ++i) aSurf[i]= 0;
  //  for (unsigned j=0; j<aDim; ++j)
  //    aSurf[i][j]= 0;
 }  

void bv_stat::set_name(char* name, char* suffix) 
     // *** changed 27.01.03 to make more general purpose
{
  if (name==NULL) aName= NULL;
  else if (suffix==NULL)
    {
      aName= new char[strlen(name)+1];
      strcpy(aName, name);
    }
  else 
    {
      aName= new char[strlen(name)+strlen(suffix)+2];
      strcpy(aName, name);
      strcat(aName, ".");
      strcat(aName, suffix);
    }
}

void bv_stat::init(char* name)
{
  set_name(name);
  zero();
}

double bv_stat::add_value(double v1, double v2)
{
  aV1= v1; aV2= v2;
  aCount++;
  int b1= (int) floor((v1 - aMinv)*aMult);
  int b2= (int) floor((v2 - aMinv)*aMult);
  int ind= b1*aDim+b2;
  if (ind<aDim*aDim) aSurf[(b1*aDim)+b2]++;
  else cerr << "bv_stat: Warning trying write outside of array\n";
  return v1;
}

float bv_stat::axis_label(int n)
{
  float rem= (aMaxv - aMinv)/(aDim*2);
  return aMinv + rem + (((float) n)/aDim)*(aMaxv -aMinv);
}

bool bv_stat::check_sum()
{  
  int sum= 0;
  for (int w=0; w<aDim; ++w)
    for (int l=0; l<aDim; ++l)
      sum+= aSurf[w*aDim + l];
  if (sum!=aCount)
    {
      cerr << "bv_stat: warning matrix sum " << sum << " does not match expected value ";
      cerr << aCount << "\n";
      return false;
    }
  else return true;
}

void bv_stat::print(bool normalise= true, bool labels= false)
{

  if (labels) {
    cout << "*\t";
    for (int w=0; w<aDim; ++w) {
      cout << axis_label(w);
      if (w==(aDim -1)) cout << endl; else cout <<'\t';
    }
  }

  for (int l=(aDim - 1); l>=0; --l)
    {
      if (labels) cout << axis_label(l) << '\t';
      for (int w=0; w<aDim; ++w)
	{
	  if (normalise) cout << ((float) aSurf[w*aDim + l])/aCount;
	  else cout << aSurf[w*aDim + l];
	  if (w<(aDim - 1)) cout << '\t'; else cout << '\n';
	}
    }
  cout << endl;
}

void bv_stat::print_cols(bool normalise= true, bool labels= false)
{
  for (int w=0; w<aDim; ++w)
    for (int l=0; l<aDim; ++l)
      {
	if (labels) cout << axis_label(w) << '\t' << axis_label(l) << '\t';
	if (normalise) cout << ((float) aSurf[w*aDim + l])/aCount;
	else cout << aSurf[w*aDim + l];
	cout << '\n';
      }
  cout << endl;
}

#endif
