/*****************************************************************************
 * File:         utils/stat.h                                                *
 * Version:      4.0                                                         *
 * Description:  Stores and calculates statistics relating to a variable     *
 * Author:       Tony Prescott                                               *
 * Contact:      T.J.Prescott@shef.ac.uk                                     *
 *****************************************************************************/

#ifndef UTILS_STAT
#define UTILS_STAT

#include <cstdlib>
// #include <stream.h>
#include <math.h>
#include <vector>
#include <algorithm>
#include <string>
#include "io.h"

typedef vector <double> db_vector;

class stat {
 private:
  char*  aName;
  int    aCount;
  int    aMax_count;
  double aValue;
  double aMin;
  double aMax;
  double aSum;
  double aSum2;
  bool   aSave;
  db_vector aData;
  bool aDistrib;
  int aNbins;
  db_vector aBins;
  double aBin_min;
  double aBin_max;
  int aStart;

  void add_data(double);
 public:
  float count() { return aCount; }
  double value() { return aValue; }
  double min() { return aMin; }
  double max() { return aMax; }
  double mean() { return aSum / aCount; }
  double median();
  double rms() { return sqrt(aSum2/aCount); }
  double sd() { return sqrt ((aSum2 - (aSum*aSum)/aCount)/(aCount - 1)); }
  double saturation() { return ((double) aMax_count)/aCount; }
  char* name() { return aName; }
  void zero();
  void set_name(char*, char* suffix= NULL);  // *** new param list 27.01.03
  void init(char*, bool);
  double add_value(double);
  void print_stats(int);
  void print(bool, int);
  void print_stats_hdr(bool, int);
  void print_stats_r(bool, int);
  void print_distrib(int);
  void print_distrib_hdr(bool, int);
  void print_distrib_r(bool, int);
  stat(char*, bool, bool, double, double);
  ~stat() { if (aName!=NULL) delete[] aName; aData.clear(); }
};

stat::stat(char* name= NULL, bool save= true, bool get_dist= false, 
	   double bmin= 0.0, double bmax= 1.0)
{
  set_name(name);
  // cout << "new stat '" << name << "' save ";
  // if (save) cout << "enabled\n";
  // else cout << "disabled\n";
  aSave= save;
  aDistrib= get_dist;
  aNbins= 10;
  aBin_min= bmin;
  aBin_max= bmax;
  zero();
}

void stat::zero() 
{
  // cout << "initialising " << aName << "\n";
  aValue= 0.0;
  aCount= 0;
  aStart= 0;
  aMin= DBL_MAX;
  aMax= DBL_MIN;
  aMax_count= 0;
  aSum= aSum2= 0.0;
  aData.clear();
  db_vector b(aNbins);    // new vector all elements 0
  aBins= b;
  // print_distrib(1);
}  

void stat::set_name(char* name, char* suffix) 
     // *** changed 27.01.03 to make more general purpose
{
  if (name==NULL) aName= NULL;
  else if (suffix==NULL)
    {
      aName= new char[strlen(name)+1];
      strcpy(aName, name);
    }
  else 
    {
      aName= new char[strlen(name)+strlen(suffix)+2];
      strcpy(aName, name);
      strcat(aName, ".");
      strcat(aName, suffix);
    }
}

void stat::init(char* name, bool save= true)
{
  set_name(name);
  aSave= save;
  zero();
}

double stat::add_value(double v)
{
  aValue= v;
  // cout << "new value " << aName << " " << v << "\n";
  if (aSave) aData.push_back(v);
  aCount++;
  if (v > aMax) { aMax= v; aMax_count= 0; }
  else if (v==aMax) aMax_count++;
  if (v < aMin) aMin= v;
  aSum += v;
  aSum2 += v*v;
  if (aDistrib) {
    if (v<aBin_min) aBins[0]++;
    else if (v>=aBin_max) aBins[aNbins-1]++;
    else {
      int b= (int) floor((v - aBin_min)*aNbins/(aBin_max - aBin_min));
      if (b>=0 && b<aNbins) aBins[b]++;
      else cerr << "stat: wrong bin!!!\n\n";
    }
  }
  return v;
}

double stat::median()
{
  if (!aSave) return -1;
  else
    {
      db_vector sorted_data(aData);
      
      sort(sorted_data.begin(), sorted_data.end());
      
      int sz= sorted_data.size();
      if (sz % 2==1)
	return sorted_data[(sz - 1)/2];
      else
	{
	  int m2= sz/2, m1= m2 - 1;
	  return (sorted_data[m1] + sorted_data[m2])/2.0;
	}
    }
}

void stat::print_distrib(int verbose= 1)
{
  if (!aDistrib) cout << "no distribution saved\n";
  else {
    cout << "distribution:\n";
    float inc= (aBin_max - aBin_min)/aNbins;
    for (unsigned i= 0; i<aBins.size(); ++i)
      cout << i*inc << '\t' << aBins[i]/aCount << "\n";
  }
}

void stat::print_stats(int verbose= 1) 
{
  if (aCount==0)
    cout << aName << ": 0 observations.\n";
  else
    {
      cout << aName << "(" << aCount << "): ";
      cout << "mn " << mean() << ", sd " << sd();
      if (aSave) cout << ", median " << median(); 
      cout << ", max " << aMax << ", min " << aMin;
      // cout << ", saturation " << 100.0*aMax_count/aCount << "%";
      cout << "\n";
      if (verbose>1) print_distrib(1);
    }
}

void stat::print_distrib_hdr(bool el=false, int verbose= 1)
{
  if (aDistrib) {
    double inc= (aBin_max - aBin_min)/aNbins;
    for (unsigned i= 0; i<(aBins.size() - 1); ++i) cout << "b_" << i*inc << "\t";
    cout << "b_" << inc*(aNbins - 1);
    if (el) cout << endl; else cout << '\t';
  }
}

void stat::print_distrib_r(bool el= false, int verbose= 1)
{
  if (aDistrib) {
    for (unsigned i= 0; i<(aBins.size() - 1); ++i) cout << aBins[i]/aCount << "\t";
    cout << aBins[aNbins - 1]/aCount;
    if (el) cout << endl; else cout << '\t';
  }
}
void stat::print_stats_hdr(bool el= false, int verbose= 0)
{
  string ss= abb_cstring(aName, 3); 
  cout << ss << "_mn";
  if (verbose>0){
    cout << '\t' << ss << "sd\t"; // << ss << "mdn\t";
    cout << ss << "max\t" << ss << "min"; // << ss << "sat";
  }
  if (verbose>1) { cout << "\t"; print_distrib_hdr(el, verbose); }
  else if (el) cout << endl; 
  else cout << '\t';
}

void stat::print_stats_r(bool el= false, int verbose= 0)
{
  cout << mean();
  if (verbose>0) {
    cout << '\t' << sd(); // << '\t' << median();
    cout<< '\t' << aMax << '\t' << aMin; //  << '\t' << saturation();
  }
  if (verbose>1) { cout << '\t'; print_distrib_r(el, verbose); }
  else if (el) cout << endl; 
  else cout << '\t';
}

void stat::print(bool pdata= false, int ndata= 0)
{
  print_stats();
  if (pdata && !aData.empty())
    {
      // sort(aData.begin(), aData.end());
      if (ndata!=0) cout << "last " << ndata << " observations:\n";
      else cout << "all data:\n";

      unsigned start= (ndata==0)? 0 : aData.size() - ndata;
      for (unsigned i= start; i<aData.size(); ++i)
	{
	  cout << i + 1 - start << ": " << aData[i] << endl;
	}
      cout << endl;
    }
}

#endif





















